#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from weakref import WeakKeyDictionary, ref

from traits.api import TraitType, TraitError, BaseInstance
from traits.traits import CTrait


#------------------------------------------------------------------------------
# Enaml Instance
#------------------------------------------------------------------------------
class EnamlInstance(TraitType):
    """ A custom TraitType which serves as a simple isinstance(...)
    validator. This class serves as the base class for other custom
    trait types such as EnamlEvent and UserAttribute.

    """
    @staticmethod
    def is_valid_type(obj):
        """ A static method which returns whether or not the given object
        can be used as the type in an isinstance(..., type) expression.

        Paramters
        ---------
        obj : object
            The object which should behave like a type for the purpose
            of an isinstance check. This means the object is type
            or defines an '__instancecheck__' method.

        Returns
        -------
        result : bool
            True if the object is a type or defines a method named
            '__instancecheck__', False otherwise.

        """
        return isinstance(obj, type) or hasattr(obj, '__instancecheck__')

    def __init__(self, base_type=object):
        """ Initialize a UserAttribute instance.

        Parameters
        ----------
        base_type : type-like object, optional
            An object that behaves like a type for the purposes of a
            call to isinstance. The staticmethod 'is_valid_attr_type'
            defined on this class can be used to test a type before
            creating an instance of this class. It is assumed that the
            given type passes that test. The default is object.

        """
        if not EnamlInstance.is_valid_type(base_type):
            msg = '%s is not a valid type object'
            raise TypeError(msg % base_type)
        super(EnamlInstance, self).__init__()
        self.base_type = base_type

    def validate(self, obj, name, value):
        """ The validation handler for an EnamlInstace. It performs a 
        simple isinstance(...) check using the attribute type provided 
        to the constructor.

        """
        if not isinstance(value, self.base_type):
            self.error(obj, name, value)
        return value

    def full_info(self, obj, name, value):
        """ Overridden parent class method to compute an appropriate info
        string for use in error messages.

        """
        return 'an instance of %s' % self.base_type


#------------------------------------------------------------------------------
# Enaml Event
#------------------------------------------------------------------------------
class EnamlEventDispatcher(object):
    """ A thin object which is used to dispatch a notification for an
    EnamlEvent. Instances of this class are callable with at most one
    argument, which will be the payload of the event. Instances of this
    dispatcher should not be held onto, since they maintain a strong 
    reference to the underlying object.

    """
    def __init__(self, trait, obj, name):
        """ Initialize an event dispatcher.

        Parameters
        ----------
        trait : Instance(TraitType)
            The trait type instance on which validate will be called
            with the event payload.
        
        obj : Instance(HasTraits)
            The HasTraits object on which the event is being emitted.
        
        name : string
            The name of the event being emitted.
        
        """
        self._trait = trait
        self._obj = obj
        self._name = name
    
    def __call__(self, payload=None):
        """ Dispatches the event with the given payload.

        Paramters
        ---------
        payload : object, optional
            The payload argument of the event. This object will be 
            validated against the type declared for the event.
            The default payload is None.
        
        """
        obj = self._obj
        name = self._name
        self._trait.validate(obj, name, payload)
        obj.trait_property_changed(name, None, payload)


class EnamlEvent(EnamlInstance):
    """ A custom EnamlInstance that is used to implement the event
    type in Enaml. An EnamlEvent is read-only, and returns a dispatcher
    which can be called to emit the event.

    """
    def get(self, obj, name):
        """ The trait getter method. Returns an EnamlEventDispatcher
        instance which can be called to emit the event.

        """
        return EnamlEventDispatcher(self, obj, name)

    def set(self, obj, name, value):
        """ The trait setter method. Fires off a EnamlEventDispatcher
        as if the event were called with the payload.

        """
        EnamlEventDispatcher(self, obj, name)(value)

    def full_info(self, obj, name, value):
        """ Overridden parent class method to compute an appropriate info
        string for use in error messages.

        """
        return 'emitted with an object of %s' % self.base_type


#------------------------------------------------------------------------------
# Expression Trait
#------------------------------------------------------------------------------
class ExpressionInitializationError(Exception):
    """ An exception used to indicate an error during initialization
    of an expression.

    """
    # XXX - We can't inherit from AttributeError because the local 
    # scope object used by expressions captures an AttributeError
    # and converts it into in a KeyError in order to implement
    # dynamic attribute scoping. We actually want this exception
    # to propagate.
    pass


class ExpressionTrait(TraitType):
    """ A custom trait type which is used to help implement expression 
    binding. Instances of this trait are added to an object, but swap
    themselves out and replace the old trait the first time they are
    accessed. This allows bound expressions to be initialized in the
    proper order without requiring an explicit initialization graph.

    """
    def __init__(self, old_trait):
        """ Initialize an expression trait.

        Parameters
        ----------
        old_trait : ctrait
            The trait object that the expression trait is temporarily
            replacing. When a 'get' or 'set' is triggered on this 
            trait, the old trait will be restored and then the default
            value of the expression will be applied.
        
        """
        super(ExpressionTrait, self).__init__()
        self.old_trait = old_trait
    
    def swapout(self, obj, name):
        """ Restore the old trait onto the object. This method takes
        care to make sure that listeners are copied over properly.

        """
        # The default behavior of add_trait does *almost* the right
        # thing when it copies over notifiers when replacing the
        # existing trait. What it fails to do is update the owner
        # attribute of TraitChangeNotifyWrappers which are managing
        # a bound method notifier. This means that if said notifier
        # ever dies, it removes itself from the incorrect owner list
        # and it will be (erroneously) called on the next dispatch
        # cycle. The logic here makes sure that the owner attribute
        # of such a notifier is properly updated with its new owner.
        obj.add_trait(name, self.old_trait)
        notifiers = obj.trait(name)._notifiers(0)
        if notifiers is not None:
            for notifier in notifiers:
                if hasattr(notifier, 'owner'):
                    notifier.owner = notifiers

    def compute_default(self, obj, name):
        """ Returns the default value as computed by the most recently
        bound expression. If a value cannot be provided, NotImplemented
        is returned.

        """
        res = NotImplemented
        expr = obj._expressions[name][0]
        if expr is not None:
            try:
                res = expr.eval()
            except Exception as e:
                # Reraise a propagating initialization error.
                if isinstance(e, ExpressionInitializationError):
                    raise
                msg = ('Error initializing expression (%r line %s). '
                       'Orignal exception was:\n%s')
                import traceback
                tb = traceback.format_exc()
                filename = expr.code.co_filename
                lineno = expr.code.co_firstlineno
                args = (filename, lineno, tb)
                raise ExpressionInitializationError(msg % args)
        return res

    def get(self, obj, name):
        """ Handle computing the initial value for the expression trait. 
        This method first restores the old trait, then evaluates the 
        expression and sets the value on the trait. It then performs
        a getattr to return the new value of the trait. If the object
        is not yet fully initialized, the value is set quietly.

        """
        self.swapout(obj, name)
        val = self.compute_default(obj, name)
        if val is not NotImplemented:
            if not obj.initialized:
                obj.trait_setq(**{name: val})
            else:    
                setattr(obj, name, val)
        return getattr(obj, name, val)

    def set(self, obj, name, val):
        """ Handle the setting of an initial value for the expression
        trait. This method first restores the old trait, then sets
        the value on that trait. In this case, the expression object
        is not needed. If the object is not yet fully initialized, the 
        value is set quietly.

        """
        self.swapout(obj, name)
        if not obj.initialized:
            obj.trait_setq(**{name: val})
        else:
            setattr(obj, name, val)


#------------------------------------------------------------------------------
# User Attribute and Event
#------------------------------------------------------------------------------
class UninitializedAttributeError(Exception):
    """ A custom Exception used by UserAttribute to signal the access 
    of an uninitialized attribute.

    """
    # XXX - We can't inherit from AttributeError because the local 
    # scope object used by expressions captures an AttributeError
    # and converts it into in a KeyError in order to implement
    # dynamic attribute scoping. We actually want this exception
    # to propagate.
    pass


class UserAttribute(EnamlInstance):
    """ An EnamlInstance subclass that is used to implement optional 
    attribute typing when adding a new user attribute to an Enaml 
    component.

    """
    def get(self, obj, name):
        """ The trait getter method. Returns the value from the object's
        dict, or raises an uninitialized error if the value doesn't exist.

        """
        dct = obj.__dict__
        if name not in dct:
            self.uninitialized_error(obj, name)
        return dct[name]

    def set(self, obj, name, value):
        """ The trait setter method. Sets the value in the object's 
        dict if it is valid, and emits a change notification if the
        value has changed. The first time the value is set the change
        notification will carry None as the old value.

        """
        value = self.validate(obj, name, value)
        dct = obj.__dict__
        if name not in dct:
            old = None
        else:
            old = dct[name]
        dct[name] = value
        if old != value:
            obj.trait_property_changed(name, old, value)

    def uninitialized_error(self, obj, name):
        """ A method which raises an UninitializedAttributeError for
        the given object and attribute name

        """
        msg = "Cannot access the uninitialized '%s' attribute of the %s object"
        raise UninitializedAttributeError(msg % (name, obj))


class UserEvent(EnamlEvent):
    """ A simple EnamlEvent subclass used to distinguish between events
    declared by the framework, and events declared by the user.

    """
    pass


#------------------------------------------------------------------------------
# Bounded
#------------------------------------------------------------------------------
class Bounded(TraitType):
    """ Generic Bounded Trait class.

    The class defines a generic Trait where the value is validated
    on assigment to fall between low and high (static or dynamic) bounds.

    """
    info_text = "Bounded value"

    def __init__(self, value=None, low=None, high=None, **metadata):
        """
        Arguments
        ---------
        value :
            The default value. It can be a python object or a Trait.

        low :
            The lower bound of the Trait.

        high :
            The upper bound of the Trait.

        """
        if isinstance(value, CTrait):
            self.validate = self.validate_with_trait
            default_value = value.default
        else:
            self.validate = self.validate_bounds
            if value is None:
                if low is not None:
                    value = default_value = low
                else:
                    value = default_value = high
            else:
                default_value = value

        super(Bounded, self).__init__(default_value, **metadata)

        self._high = high
        self._low = low
        self._value = value

        if isinstance(value, basestring):
            self.default_value_type = 8
            self.default_value = self._get_default_value
    
    def _get_default_value(self, obj):
        """ Handles computing the default value for the Bounded trait.

        """
        return reduce(getattr, self._value.split('.'), obj)

    def validate_with_trait(self, obj, name, value):
        """ Validate the trait value.

        Validation takes place in two steps:
        #. The input value is validated based on the expected Trait type.
        #. The value it is between the static (or dynamic) bounds.

        """
        value_trait = self._value
        value = value_trait.validate(obj, name, value)
        return self.validate_bounds(obj, name, value)

    def validate_bounds(self, obj, name, value):
        """ Validate that the value is in range.

        .. note:: Any exceptions that may take place are converted to
            TraitErrors.

        """
        low, high = self.get_bounds(obj)
        if low is None:
            low = value
        if high is None:
            high = value
        
        is_inside_bounds = False
        try:
            is_inside_bounds = (low <= value <= high)
        except Exception as raised_exception:
            if isinstance(raised_exception, TraitError):
                raise raised_exception
            else:
                msg = ('Bound checking of {0} caused a the following Python '
                       'Exception: {1}'.format(value, raised_exception))
                raise TraitError(msg)
        
        if not is_inside_bounds:
            msg = ('The assigned date value must be bounded between {0} '
                   ' and {1}. Got {2} instead.'.format(low, high, value))
            raise TraitError(msg)

        return value

    def get_bounds(self, obj):
        """ Get the lower and upper bounds of the Trait.

        """
        low = self._low
        if isinstance(low, basestring):
            low = reduce(getattr, low.split('.'), obj)
        
        high = self._high
        if isinstance(high, basestring):
            high = reduce(getattr, high.split('.'), obj)

        return (low, high)


#------------------------------------------------------------------------------
# Lazy Property
#------------------------------------------------------------------------------
class LazyProperty(TraitType):
    """ A trait which behaves like a read-only cached property, but
    which lazily defers binding the dependency notifiers until the
    first time the value is retrieved. It is used to avoid situations
    where a property dependency in prematurely evaluated during
    component instantiation.

    """
    def __init__(self, trait=None, depends_on=''):
        """ Initialize a LazyProperty.

        Parameters
        ----------
        trait : TraitType, optional
            An optional trait type for the values returned by the
            property. List is required if using extending trait 
            name syntax for e.g. list listeners.
        
        depends_on : string, optional
            The traits notification string for the dependencies of
            the filter.
        
        """
        super(LazyProperty, self).__init__()
        self.dependency = depends_on
        self.handlers = WeakKeyDictionary()
        if trait is not None:
            self.default_value_type = trait.default_value_type

    def get(self, obj, name):
        """ Returns the (possibly cached) value of the filter. The 
        notification handlers will be attached the first time the
        value is accessed.

        """
        cache_name = '_%s_lazy_property_cache' % name
        dct = obj.__dict__
        if cache_name not in dct:
            method_name = '_get_%s' % name
            val = getattr(obj, method_name)()
            dct[cache_name] = val
            self.bind(obj, name)
        else:
            val = dct[cache_name]
        return val
    
    def bind(self, obj, name):
        """ Binds the dependency notification handlers for the object.

        """
        wr_obj = ref(obj)
        def notify():
            obj = wr_obj()
            if obj is not None:
                cache_name = '_%s_lazy_property_cache' % name
                old = obj.__dict__.pop(cache_name, None)
                obj.trait_property_changed(name, old)

        handlers = self.handlers
        dependency = self.dependency
        if obj in handlers:
            obj.on_trait_change(handlers[obj], dependency, remove=True)
        handlers[obj] = notify
        obj.on_trait_change(notify, dependency)


#------------------------------------------------------------------------------
# Coercing Instance
#------------------------------------------------------------------------------
class CoercingInstance(BaseInstance):
    """ A BaseInstance subclass which attempts to coerce a value by 
    calling the class constructor and passing the new value into the
    original validate method.

    """
    def validate(self, obj, name, value):
        """ Attempts to coerce the given value to an appropriate type
        by calling the underlying class constructor. The coerced value
        is then sent to the superclass' validate method.

        """
        if isinstance(self.klass, basestring):
            self.resolve_klass(obj, name, value)
        try:
            value = self.klass(value)
        except:
            pass
        return super(CoercingInstance, self).validate(obj, name, value)


#------------------------------------------------------------------------------
# Enaml Widget Instance
#------------------------------------------------------------------------------
class EnamlWidgetInstance(BaseInstance):
    """ A BaseInstance subclass which will call the 'setup' method on
    the value being assigned, passing in the appropriate toolkit widget
    object. The 'destroy' method on the old instance will be called 
    unless the 'destroy_old' flag is set to False. These operations are
    done *before* the value is actually set and therefore before any 
    change notifications are fired.

    """
    def __init__(self, *args, **kwargs):
        self.destroy_old = kwargs.pop('destroy_old', True)
        super(EnamlWidgetInstance, self).__init__(*args, **kwargs)

    def validate(self, obj, name, value):
        """ Validates the value using the superclass method. Upon
        success, it then destroys the old value (if requested) and 
        sets up the new value. The value is expected to be at least
        an instance WidgetComponent.

        """
        value = super(EnamlWidgetInstance, self).validate(obj, name, value)
        old = getattr(obj, name, None)
        if old is not None and self.destroy_old:
            old.destroy()
        value.setup(obj.toolkit_widget)
        return value




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt_container import QtContainer

from ...components.tab import AbstractTkTab


class QtTab(QtContainer, AbstractTkTab):
    """ A Qt implementation of the Tab component.
    
    """
    #--------------------------------------------------------------------------
    # Setup Methods 
    #--------------------------------------------------------------------------
    def initialize(self):
        """ Initialize the attributes of the tab.

        """
        super(QtTab, self).initialize()
        shell = self.shell_obj
        self.set_title(shell.title)
        self.set_icon(shell.icon)

    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_title_changed(self, title):
        """ The change handler for the 'title' attribute on the shell 
        object.

        """
        self.set_title(title)
    
    def shell_icon_changed(self, icon):
        """ The change handler for the 'icon' attribute on the shell
        object.

        """
        self.set_icon(icon)
    
    #--------------------------------------------------------------------------
    # Widget Update Methods 
    #--------------------------------------------------------------------------
    def set_title(self, title):
        """ Sets the title of this tab in the parent tab widget.

        """
        # We use got through the shell's parent object to get at the
        # tab widget for two reasons: 1) The TabWidget maintains an 
        # internal QStackedWidget control which is actually the parent 
        # of the tab, and 2) If the Tab was generated by an include
        # then the parents end up in a transitory state where logic
        # can break down. Accessing the tab widget through the parent
        # shell is therefore the safest option.
        widget = self.widget
        tab_widget = self.shell_obj.parent.toolkit_widget
        idx = tab_widget.indexOf(widget)
        if idx != -1:
            tab_widget.setTabText(idx, title)
     
    def set_icon(self, icon):
        """ Sets the icon of this tab in the parent tab widget.

        """
        # XXX handle icons
        pass




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import abstractmethod

from traits.api import Instance, Bool

from .abstract_item_view import AbstractItemView, AbstractTkItemView


class AbstractTkTableView(AbstractTkItemView):
    
    @abstractmethod
    def shell_vertical_header_visible_changed(self, visible):
        raise NotImplementedError
    
    @abstractmethod
    def shell_horizontal_header_visible_changed(self, visible):
        raise NotImplementedError


class TableView(AbstractItemView):
    """ A view for tabular data.
    
    """
    #: Whether or not the vertical header is shown. Defaults to True.
    vertical_header_visible = Bool(True)

    #: Whether or not the horizontal header is shown. Defaults to True.
    horizontal_header_visible = Bool(True)

    #: How strongly a component hugs it's contents' width. TableViews 
    #: ignore the width hug by default, so they expand freely in width.
    hug_width = 'ignore'
    
    #: How strongly a component hugs it's contents' height. TableViews 
    #: ignore the height hug by default, so they expand freely in height.
    hug_height = 'ignore'

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkTableView)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt_test_assistant import QtTestAssistant
from .. import check_box

class TestQtCheckBox(QtTestAssistant, check_box.TestCheckBox):
    """ QtCheckbox tests. """

    def get_text(self, widget):
        """ Returns the text from the tookit widget.

        """
        return widget.text()

    def get_checked(self, widget):
        """ Returns the checked status from the tookit widget.

        """
        return widget.isChecked()

    def checkbox_pressed(self, widget):
        """ Press the checkbox programmatically.

        """
        self.widget.pressed.emit()

    def checkbox_released(self, widget):
        """ Release the button programmatically.

        """
        self.widget.released.emit()

    def checkbox_toggle(self, widget):
        """ Toggle the button programmatically.

        """
        self.widget.toggled.emit(True)



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from ..core.constructor import Constructor


def include_shell():
    from .include import Include
    return Include


def inline_shell():
    from .inline import Inline
    return Inline


CONSTRUCTORS = (
    ('Include', Constructor(include_shell)),
    ('Inline', Constructor(inline_shell)),
)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx
import wx.lib.newevent

from .wx_base_widget_component import WXBaseWidgetComponent
from .styling import wx_color_from_color, wx_font_from_font

from ...components.widget_component import AbstractTkWidgetComponent
from ...layout.geometry import Rect, Size, Pos


#: An event that is emitted when the minimum size is set on a widget. 
#: This is required by the WXMainWindowSizer to know when to resize
#: the frame if the central widget's min size has changed.
wxMinSizeChanged, EVT_MIN_SIZE = wx.lib.newevent.NewEvent()


class WXWidgetComponent(WXBaseWidgetComponent, AbstractTkWidgetComponent):
    """ A Wx implementation of WidgetComponent.

    """
    def create(self, parent):
        """ Creates the underlying Wx widget. As necessary, subclasses
        should reimplement this method to create different types of
        widgets.

        """
        self.widget = wx.Panel(parent)

    def initialize(self):
        """ Initializes the attributes of the the Wx widget.

        """
        super(WXWidgetComponent, self).initialize()
        shell = self.shell_obj
        self.set_enabled(shell.enabled)
        if shell.bgcolor:
            self.set_bgcolor(shell.bgcolor)
        if shell.fgcolor:
            self.set_fgcolor(shell.fgcolor)
        if shell.font:
            self.set_font(shell.font)

    def enable_updates(self):
        """ Enable rendering updates for the underlying Wx widget.

        """
        widget = self.widget
        if widget:
            if widget.IsFrozen():
                widget.Thaw()
    
    def disable_updates(self):
        """ Disable rendering updates for the underlying Wx widget.

        """
        widget = self.widget
        if widget:
            if not widget.IsFrozen():
                widget.Freeze()

    def set_visible(self, visible):
        """ Show or hide the widget.

        """
        self.widget.Show(visible)

    def size_hint(self):
        """ Returns a (width, height) tuple of integers which represent
        the suggested size of the widget for its current state, ignoring
        any windowing decorations. This value is used by the layout 
        manager to determine how much space to allocate the widget.

        """
        return Size(*self.widget.GetBestSizeTuple())

    def layout_geometry(self):
        """ Returns the (x, y, width, height) to of layout geometry
        info for the internal toolkit widget. This should ignore any
        windowing decorations, and may be different than the value
        returned by geometry() if the widget's effective layout rect
        is different from its paintable rect.

        """
        return self.geometry()

    def set_layout_geometry(self, rect):
        """ Sets the layout geometry of the internal widget to the 
        given x, y, width, and height values. The parameters passed
        are equivalent semantics to layout_geometry().

        """
        self.set_geometry(rect)

    def geometry(self):
        """ Returns an (x, y, width, height) tuple of geometry info
        for the internal toolkit widget, ignoring any windowing
        decorations.

        """
        # wx widget.GetClientRect() doesn't seem to be what we want.
        # Often, GetRect() and GetClientRect() are the same, but when
        # they aren't, GetRect() does the right thing. This is probably
        # because of asymmetry between GetClientRect and SetDimensions. 
        # Given that, the existing examples have been tested side by 
        # side with Qt on Windows, and these current geometry handlers 
        # "do the right thing" in most cases, plus or minus a pixel or
        # two difference.
        return Rect(*self.widget.GetRect().asTuple())

    def set_geometry(self, rect):
        """ Sets the geometry of the internal widget to the given
        x, y, width, and height values, ignoring any windowing 
        decorations.

        """
        # wx doesn't seem to have a SetClientDimensions function
        # or anything similar. SetClientRect *does not* do the right
        # thing. This SetDimensions *seems* to do the correct thing
        # but I'm skeptical that it won't break down in certain 
        # cirumstances and do the wrong thing wrt to window decorations.
        self.widget.SetDimensions(*rect)

    def min_size(self):
        """ Returns the hard minimum (width, height) of the widget, 
        ignoring any windowing decorations. A widget will not be able
        to be resized smaller than this value

        """
        widget = self.widget
        widget_width, widget_height = widget.GetSizeTuple()
        client_width, client_height = widget.GetClientSizeTuple()
        min_width, min_height = tuple(widget.GetMinSize())
        min_width = min_width - (widget_width - client_width)
        min_height = min_height - (widget_height - client_height)
        return Size(min_width, min_height)

    def set_min_size(self, size):
        """ Set the hard minimum width and height of the widget, ignoring
        any windowing decorations. A widget will not be able to be resized 
        smaller than this value.

        """
        # Wx is lacking a SetMinClientSize function, so we fake it.
        widget = self.widget
        min_width, min_height = size
        widget_width, widget_height = widget.GetSizeTuple()
        client_width, client_height = widget.GetClientSizeTuple()
        delta_width = widget_width - client_width
        delta_height = widget_height - client_height
        min_width = min_width + delta_width
        min_height = min_height + delta_height

        # Wx won't automatically reset the max size if the min size
        # is larger, ugh. Additionally, we don't want to update the
        # max size if it's < 0 since that indicates the max size is
        # not bounded. 
        max_width, max_height = self.max_size()
        if max_width >= 0 and max_height >= 0:
            max_width = max(min_width, max_width + delta_width)
            max_height = max(min_height, max_height + delta_height)
            self.widget.SetMaxSize((max_width, max_height))

        widget.SetMinSize((min_width, min_height))

        # If the min size is now smaller that the existing size, resize
        # the widget to meet the min size.
        #
        # XXX this is a bit wrong, it should only resize the offending
        # dimension.
        new_width = max(widget_width, min_width)
        new_height = max(widget_height, min_height)
        new_size = (new_width, new_height)
        if new_size != (widget_width, widget_height):
            widget.SetSize(new_size)

        # We create and emit a min size event so that if this widget
        # is the central widget in a MainWindow, the sizer for that
        # window can properly update the minimum size of the frame.
        evt = wxMinSizeChanged()
        wx.PostEvent(widget, evt)

    def max_size(self):
        """ Returns the hard maximum (width, height) of the widget, 
        ignoring any windowing decorations. A widget will not be able
        to be resized larger than this value

        """
        # Wx is lacking a MaxClientSize function, so we fake it.
        widget = self.widget
        widget_width, widget_height = widget.GetSizeTuple()
        client_width, client_height = widget.GetClientSizeTuple()
        max_width, max_height = tuple(widget.GetMaxSize())
        max_width = max_width - (widget_width - client_width)
        max_height = max_height - (widget_height - client_height)
        return Size(max_width, max_height)

    def set_max_size(self, size):
        """ Set the hard maximum width and height of the widget, ignoring
        any windowing decorations. A widget will not be able to be resized 
        larger than this value.

        """
        # Wx is lacking a SetMaxClientSize function and at the moment, 
        # it's uncertain what the maximum allowable size is on wx. For 
        # now, we just fake computation and use the Qt limits.
        widget = self.widget
        max_width, max_height = size
        widget_width, widget_height = widget.GetSizeTuple()
        client_width, client_height = widget.GetClientSizeTuple()
        delta_width = widget_width - client_width
        delta_height = widget_height - client_height 
        max_width = max_width + delta_width
        max_height = max_height + delta_height
        max_width = min(max_width, 16777215)
        max_height = min(max_height, 16777215)

        # Wx won't automatically reset the min size if the max size
        # is smaller. Ugh....
        min_width, min_height = self.min_size()
        min_width = min(min_width + delta_width, max_width)
        min_height = min(min_height + delta_height, max_height)
        self.widget.SetMinSize((min_width, min_height))

        self.widget.SetMaxSize((max_width, max_height))

        # If the max size is now larger that the existing size, resize
        # the widget to meet the max size.
        #
        # XXX this is a bit wrong, it should only resize the offending
        # dimension.
        new_width = min(widget_width, max_width)
        new_height = min(widget_height, max_height)
        new_size = (new_width, new_height)
        if new_size != (widget_width, widget_height):
            widget.SetSize(new_size)

    def size(self):
        """ Returns the size of the internal toolkit widget, ignoring any
        windowing decorations, as a (width, height) tuple of integers.

        """
        return Size(*self.widget.GetClientSizeTuple())
                
    def resize(self, size):
        """ Resizes the internal toolkit widget according the given
        width and height integers, ignoring any windowing decorations.

        """
        # SetClientSize completely breaks down on Windows, so we fake 
        # it and hope for the best.
        widget = self.widget
        width, height = size
        widget_width, widget_height = widget.GetSizeTuple()
        client_width, client_height = widget.GetClientSizeTuple() 
        width = width + (widget_width - client_width)
        height = height + (widget_height - client_height)
        self.widget.SetSize((width, height))

    def pos(self):
        """ Returns the position of the internal toolkit widget as an
        (x, y) tuple of integers, including any windowing decorations. 
        The coordinates should be relative to the origin of the widget's 
        parent, or to the screen if the widget is toplevel.

        """
        return Pos(*self.widget.GetPositionTuple())
            
    def move(self, pos):
        """ Moves the internal toolkit widget according to the given
        x and y integers which are relative to the origin of the
        widget's parent and includes any windowing decorations.

        """
        self.widget.Move(pos)

    def shell_enabled_changed(self, enabled):
        """ The change handler for the 'enabled' attribute on the shell
        object.

        """
        self.set_enabled(enabled)

    def shell_bgcolor_changed(self, color):
        """ The change handler for the 'bgcolor' attribute on the shell
        object. Sets the background color of the internal widget to the 
        given color.
        
        """
        self.set_bgcolor(color)
    
    def shell_fgcolor_changed(self, color):
        """ The change handler for the 'fgcolor' attribute on the shell
        object. Sets the foreground color of the internal widget to the 
        given color.

        """
        self.set_fgcolor(color)

    def shell_font_changed(self, font):
        """ The change handler for the 'font' attribute on the shell 
        object. Sets the font of the internal widget to the given font.

        """
        self.set_font(font)

    def set_enabled(self, enabled):
        """ Enable or disable the widget.

        """
        self.widget.Enable(enabled)

    def set_bgcolor(self, color):
        """ Sets the background color of the widget to an appropriate
        wxColor given the provided Enaml Color object.

        """
        wx_color = wx_color_from_color(color)
        self.widget.SetBackgroundColour(wx_color)

    def set_fgcolor(self, color):
        """ Sets the foreground color of the widget to an appropriate
        wxColor given the provided Enaml Color object.

        """
        wx_color = wx_color_from_color(color)
        self.widget.SetForegroundColour(wx_color)

    def set_font(self, font):
        """ Sets the font of the widget to an appropriate QFont given 
        the provided Enaml Font object.

        """
        # There's no such thing as a NullFont on wx, so if the font is
        # equivalent to the Enaml default font, and we haven't yet changed 
        # the font for this instance, then we don't change it. Otherwise
        # the fonts won't be equivalelnt to the default.
        if not font and self._has_default_wx_font:
            return

        wx_font = wx_font_from_font(font)
        self.widget.SetFont(wx_font)
        self._has_default_wx_font = False




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .enaml_test_case import EnamlTestCase, required_method


class SelectionTestHelper(object):
    """ Helper mixin for selection model test cases.

    """
    @property
    def widget(self):
        """ Get the current toolkit "widget", as it may change.

        """
        return self.component.abstract_obj.selection_model

    def index(self, row, column):
        """ Create an appropriate ModelIndex.

        """
        return self.item_model.create_index(row, column, self.item_model)

    def to_enaml_selection(self, pysel):
        """ Convert a selection list given with (row, col) tuples to a 
        selection list with ModexIndexes.

        """
        esel = []
        for topleft, botright in pysel:
            esel.append((self.index(*topleft), self.index(*botright)))
        return esel

    def from_enaml_selection(self, esel):
        """ Convert an Enaml selection list with ModelIndexes to one 
        given with (row, col) tuples for comparison purposes.

        """
        pysel = []
        for topleft, botright in esel:
            pysel.append(((topleft.row, topleft.column), 
                          (botright.row, botright.column)))
        return pysel

    def set_py_selection(self, pysel, command):
        """ Set the selection using (int, int) indices.

        """
        esel = self.to_enaml_selection(pysel)
        self.component.set_selection(esel, command)

    def get_py_selection(self):
        """ Get the selection using (int, int) indices.

        """
        return self.from_enaml_selection(self.component.get_selection())

    #--------------------------------------------------------------------------
    # Abstract methods
    #--------------------------------------------------------------------------
    @required_method
    def get_tk_selection(self, widget):
        """ Return the widget's selection as a list of (topleft, botright)
        ranges with (row, col) indexes.

        """
        pass


class TestBaseSelectionModel(EnamlTestCase, SelectionTestHelper):
    """ Logic for testing selection models.

    """
    def setUp(self):
        """ Set up tests for Enaml's BaseSelectionModel.

        """
        enaml_source = """
from enaml.item_models.standard_models import TableModel
import numpy as np

nrows = 20
ncols = 10

table = np.arange(nrows * ncols).reshape((nrows, ncols))
the_item_model = TableModel(table)

enamldef MainView(MainWindow):
    attr events
    TableView:
        name = 'table_view'
        item_model = the_item_model
        BaseSelectionModel:
            name = 'selection_model'
"""

        self.events = []
        self.view = self.parse_and_create(enaml_source, events=self.events)
        self.table_view = self.component_by_name(self.view, 'table_view')
        self.component = self.table_view.selection_model
        self.item_model = self.table_view.item_model

    def test_empty_initial_selection(self):
        """ No selection.

        """
        self.assertEqual(self.get_tk_selection(self.widget), [])
        self.assertEqual(self.get_py_selection(), [])

    def test_set_selection_clear_select(self):
        """ Test the 'clear_select' command.

        """
        pysel = [((1, 2), (3, 4))]
        self.set_py_selection(pysel, 'clear_select')
        self.assertEqual(self.get_tk_selection(self.widget), pysel)
        self.assertEqual(self.get_py_selection(), pysel)

        pysel = [((0, 1), (6,7))]
        self.set_py_selection(pysel, 'clear_select')
        self.assertEqual(self.get_tk_selection(self.widget), pysel)
        self.assertEqual(self.get_py_selection(), pysel)

        self.component.clear()
        self.assertEqual(self.get_tk_selection(self.widget), [])
        self.assertEqual(self.get_py_selection(), [])

    def test_set_selection_clear_select_rows(self):
        """ Test the ('clear_select', 'rows') command.

        """
        pysel = [((1, 2), (3, 4))]
        test_sel = [((1, 0), (3, 9))]
        self.set_py_selection(pysel, ('clear_select', 'rows'))
        self.assertEqual(self.get_tk_selection(self.widget), test_sel)
        self.assertEqual(self.get_py_selection(), test_sel)

    def test_set_selection_no_update(self):
        """ Test the 'no_update' command.

        """
        pysel = [((1, 2), (3, 4))]
        self.set_py_selection(pysel, 'no_update')
        self.assertEqual(self.get_tk_selection(self.widget), [])
        self.assertEqual(self.get_py_selection(), [])

        new = [((0, 1), (6, 7))]
        self.set_py_selection(new, 'clear_select')
        self.assertEqual(self.get_tk_selection(self.widget), new)
        self.assertEqual(self.get_py_selection(), new)

        self.set_py_selection(pysel, 'no_update')
        self.assertEqual(self.get_tk_selection(self.widget), new)
        self.assertEqual(self.get_py_selection(), new)

    def test_set_selection_select(self):
        """ Test the 'select' command.

        """
        pysel = [((1, 2), (3, 4))]
        self.set_py_selection(pysel, 'select')
        self.assertEqual(self.get_tk_selection(self.widget), pysel)
        self.assertEqual(self.get_py_selection(), pysel)

        new = [((5, 1), (7, 5))]
        self.set_py_selection(new, 'select')
        self.assertEqual(self.get_tk_selection(self.widget), pysel+new)
        self.assertEqual(self.get_py_selection(), pysel+new)

    def test_set_selection_deselect(self):
        """ Test the 'deselect' command.

        """
        pysel = [((1, 2), (3, 4))]
        self.set_py_selection(pysel, 'clear_select')
        self.assertEqual(self.get_tk_selection(self.widget), pysel)
        self.assertEqual(self.get_py_selection(), pysel)

        new = [((2, 2), (4, 4))]
        remainder = [((1,2), (1,4))]
        self.set_py_selection(new, 'deselect')
        self.assertEqual(self.get_tk_selection(self.widget), remainder)
        self.assertEqual(self.get_py_selection(), remainder)

    def test_set_selection_toggle(self):
        """ Test the 'toggle' command.

        """
        pysel = [((1, 2), (3, 4))]
        self.set_py_selection(pysel, 'clear_select')
        self.assertEqual(self.get_tk_selection(self.widget), pysel)
        self.assertEqual(self.get_py_selection(), pysel)

        new = [((2, 2), (4, 4))]
        remainder = [((1, 2), (1, 4)), ((4, 2), (4, 4))]
        self.set_py_selection(new, 'toggle')
        self.assertEqual(self.get_tk_selection(self.widget), remainder)
        self.assertEqual(self.get_py_selection(), remainder)


class TestRowSelectionModel(EnamlTestCase, SelectionTestHelper):
    """ Logic for testing RowSelectionModel

    """
    def setUp(self):
        """ Set up tests for Enaml's RowSelectionModel

        """
        enaml_source = """
from enaml.item_models.standard_models import TableModel
import numpy as np

nrows = 20
ncols = 10

table = np.arange(nrows * ncols).reshape((nrows, ncols))
the_item_model = TableModel(table)


enamldef MainView(MainWindow):
    attr events
    TableView:
        name = 'table_view'
        item_model = the_item_model
        RowSelectionModel:
            name = 'selection_model'
"""

        self.events = []
        self.view = self.parse_and_create(enaml_source, events=self.events)
        self.table_view = self.component_by_name(self.view, 'table_view')
        self.component = self.table_view.selection_model
        self.item_model = self.table_view.item_model

    def test_set_selected_rows(self):
        """ Test the selection of rows through the selected_rows trait.

        """
        self.assertEqual(self.component.selected_rows, [])
        self.component.selected_rows = [2, 3, 5, 6]
        pysel = [((2, 0), (2, 9)), ((3, 0), (3, 9)), ((5, 0), (5, 9)), ((6, 0), (6, 9))]
        self.assertEqual(self.get_tk_selection(self.widget), pysel)
        self.assertEqual(self.get_py_selection(), pysel)

        self.component.selected_rows.append(7)
        new = pysel + [((7, 0), (7, 9))]
        self.assertEqual(self.get_tk_selection(self.widget), new)
        self.assertEqual(self.get_py_selection(), new)

        del self.component.selected_rows[1]
        del new[1]
        self.assertEqual(self.get_tk_selection(self.widget), new)
        self.assertEqual(self.get_py_selection(), new)

    def test_get_selected_rows(self):
        """ Test that the selected_rows trait gets updated correctly when the
        selection is set elsewhere.

        """
        pysel = [((2, 0), (2, 9)), ((3, 0), (3, 9)), ((5, 0), (5, 9)), ((6, 0), (6, 9))]
        self.set_py_selection(pysel, ('clear_select', 'rows'))
        self.assertEqual(self.component.selected_rows, [2, 3, 5, 6])




#------------------------------------------------------------------------------
# Copyright (c) 2011, Enthought, Inc.
# All rights reserved.
#------------------------------------------------------------------------------
from .qt.QtGui import QDockWidget
from .qt.QtCore import Qt, Signal
from .qt_widget_component import QtWidgetComponent
from .noncomponents.qt_dock_manager import DOCK_AREA_MAP

from ...components.dock_pane import AbstractTkDockPane


#: A mapping from Qt dock area enum values to Enaml dock area values.
DOCK_AREA_INVERSE_MAP = dict(
    (value, key) for (key, value) in DOCK_AREA_MAP.iteritems()
)


class QtDockWidget(QDockWidget):
    """ A QDockWidget subclass which converts a close event into a 
    closed signal.

    """
    closed = Signal()

    def closeEvent(self, event):
        super(QtDockWidget, self).closeEvent(event)
        self.closed.emit()


class QtDockPane(QtWidgetComponent, AbstractTkDockPane):
    """ A Qt4 implementation of DockPane.

    """
    #--------------------------------------------------------------------------
    # Setup Methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying QDockWidget instance.

        """
        self.widget = QtDockWidget(parent)

    def initialize(self):
        """ Initializes the attributes of the QDockWidget.

        """
        super(QtDockPane, self).initialize()
        shell = self.shell_obj
        self.set_title(shell.title)
        self.set_closable(shell.closable)
        self.set_movable(shell.movable)
        self.set_floatable(shell.floatable)
        self.set_floating(shell.floating)
        self.set_title_bar_orientation(shell.title_bar_orientation)
        self.set_allowed_dock_areas(shell.allowed_dock_areas)
        self.set_dock_widget(shell.dock_widget)

    def bind(self):
        """ Binds the signal handlers for the QDockWidget.

        """
        super(QtDockPane, self).bind()
        widget = self.widget
        widget.dockLocationChanged.connect(self.on_dock_location_changed)
        widget.topLevelChanged.connect(self.on_top_level_changed)
        widget.closed.connect(self.on_close)

    #--------------------------------------------------------------------------
    # Shell Object Change Handlers
    #--------------------------------------------------------------------------
    def shell_title_changed(self, title):
        """ Updates the title bar text in the dock pane.

        """
        self.set_title(title)
            
    def shell_closable_changed(self, closable):
        """ Updates whether or not the dock pane should have a close
        button in the title bar.

        """
        self.set_closable(closable)
            
    def shell_movable_changed(self, movable):
        """ Updates whether or not the dock pane should be movable from
        its current position.

        """
        self.set_movable(movable)
            
    def shell_floatable_changed(self, floatable):
        """ Updates whether or not the dock pane is allowed to be 
        undocked an floated over the main window.

        """
        self.set_floatable(floatable)

    def shell_floating_changed(self, floating):
        """ Updates whether or no the dock pane is floating over
        the main window instead of docked in its area.

        """
        self.set_floating(floating)

    def shell_title_bar_orientation_changed(self, orientation):
        """ Updates the orientation of the title bar.

        """
        self.set_title_bar_orientation(orientation)
        
    def shell_allowed_dock_areas_changed(self, allowed_areas):
        """ Updates the allowable docking areas for the dock pane.

        """
        self.set_allowed_dock_areas(allowed_areas)
            
    def shell_dock_widget_changed(self, dock_widget):
        """ Updates the widget being managed by the dock pane.

        """
        self.set_dock_widget(dock_widget)

    #--------------------------------------------------------------------------
    # Signal Handlers
    #--------------------------------------------------------------------------
    def on_dock_location_changed(self, new_qarea):
        """ The signal handler for the 'dockLocationChanged' signal. It
        updates the 'dock_area' attribute on the shell object and fires
        off the 'moved' event.

        """
        shell = self.shell_obj
        old_area = shell.dock_area
        new_area = DOCK_AREA_INVERSE_MAP[new_qarea]
        shell.dock_area = new_area
        shell.moved((old_area, new_area))

    def on_top_level_changed(self, floating):
        """ The signal handler for the 'topLevelChanged' signal. It 
        fires of the 'moved' event and the appropriate 'docked' or
        'undocked' event on the shell object.

        """
        shell = self.shell_obj
        shell.floating = floating
        if floating:
            shell.undocked()
        else:
            shell.docked()

    def on_close(self):
        """ The signal handler for the 'closed' signal. It sets the
        visibility of the shell component to False.

        """
        self.shell_obj.visible = False

    #--------------------------------------------------------------------------
    # Widget Update Methods
    #--------------------------------------------------------------------------
    def set_title(self, title):
        """ Sets the title bar text of the dock widget.

        """
        self.widget.setWindowTitle(title)

    def set_closable(self, closable):
        """ Sets the closable flag of the title bar.

        """
        flag = QDockWidget.DockWidgetClosable
        self._update_feature_flag(flag, closable)
    
    def set_movable(self, movable):
        """ Sets the movable flag of the title bar.

        """
        flag = QDockWidget.DockWidgetMovable
        self._update_feature_flag(flag, movable)
    
    def set_floatable(self, floatable):
        """ Sets the floatable flag of the title bar.

        """
        flag = QDockWidget.DockWidgetFloatable
        self._update_feature_flag(flag, floatable)

    def set_floating(self, floating):
        """ Sets the whether or not the dock widget is floating.

        """
        widget = self.widget
        if widget.isFloating() != floating:
            widget.setFloating(floating)

    def set_title_bar_orientation(self, orientation):
        """ Sets the orientation flag of the title bar.

        """
        flag = QDockWidget.DockWidgetVerticalTitleBar
        self._update_feature_flag(flag, orientation == 'vertical')

    def set_allowed_dock_areas(self, allowed_areas):
        """ Sets the flags for the allowed docking areas.

        """
        flags = Qt.NoDockWidgetArea
        for area in allowed_areas:
            flags |= DOCK_AREA_MAP[area]
        self.widget.setAllowedAreas(flags)

    def set_dock_widget(self, dock_widget):
        """ Sets the widget being managed by the dock widget.

        """
        self.widget.setWidget(dock_widget.toolkit_widget)
    
    #--------------------------------------------------------------------------
    # Auxiliary Methods
    #--------------------------------------------------------------------------
    def _update_feature_flag(self, flag, add=True):
        """ Updates the feature flag for the dock widget.

        Parameters
        ----------
        flag : QDockWidget feature flag
            The feature flag to apply to the dock widget.
        
        add : bool, optional
            Whether to add or remove the flag from the dock widget 
            features. The default is True.
        
        """
        widget = self.widget
        features = widget.features()
        if add:
            features |= flag
        else:
            features &= ~flag
        widget.setFeatures(features)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod
from weakref import ref

from traits.api import HasTraits, Disallow

from .byteplay import (
    CALL_FUNCTION, ROT_THREE, LOAD_CONST, LOAD_ATTR, ROT_TWO, BUILD_TUPLE, 
    UNPACK_SEQUENCE, POP_TOP, DUP_TOP,
)
from .signaling import Signal


#------------------------------------------------------------------------------
# Abstract Monitor
#------------------------------------------------------------------------------
class AbstractMonitor(object):
    """ An abstract base class which defines the api for implementing 
    an expression monitor. 

    An expression Monitor is responsible for generating bytecode which
    will be inserted into a Python expression and which inspects the
    Python stack during the evaluation to determine if it is appropriate
    to attach some form of notifier to the constiuents. If a notifier is 
    warranted, it should make the appropriate connections to emit the
    expression_changed signal when the expression has changed.

    """
    __metaclass__ = ABCMeta

    #: A signal which is emitted when the expression has changed.
    expression_changed = Signal()

    @abstractmethod
    def get_insertion_code(self, code_list):
        """ Generates the byteplay code operations to be inserted into 
        the expression code object in order to monitor execution.

        Parameters
        ----------
        code_list : list of (op_code, op_arg)
            The list of byteplay code operations for the expression.
            If no code need be generated, an empty list should be 
            returned.

        Returns
        -------
        result : list of (insertion_idx, code_ops)
            A list of 2-tuples where the first item is an integer index
            into the original code_list and the second item is the list
            of byteplay code operations to insert into the final code.

        Notes
        -----
        The generated instertion code *must* have a net-zero effect on 
        the Python stack. This means that the inserted code should leave
        the stack exactly the way it found it. If this is not maintained,
        then random exceptions and/or crashes *will* result.

        """
        raise NotImplementedError

    @abstractmethod
    def reset(self):
        """ Unhook any previously connected notifiers. This method is 
        called by the owner expression when notifiers should be removed.

        """
        raise NotImplementedError


#------------------------------------------------------------------------------
# Abstract Attribute Monitor
#------------------------------------------------------------------------------
class AbstractAttributeMonitor(AbstractMonitor):
    """ An abstract monitor which monitors the expression evaluation for 
    attribute access and calls a method with the object and attribute 
    which is being accessed.
    
    """
    def get_insertion_code(self, code_list):
        """ Generates the byteplay code operations to be inserted into 
        the expression code object to monitor attribute accesses. When
        an attribute access occurs, the 'monitor_attribute' method will 
        be called with the object and attribute name as arguments.

        """
        # A weak closure which is injected into the stack to call the
        # monitor method upon attribute access.
        def code_binder(obj, attr, selfref=ref(self)):
            this = selfref()
            if this is not None:
                this.monitor_attribute(obj, attr)
        
        # The list of code segments that will be inserted into the
        # new bytecode for the expression.
        insertion_code = []

        for idx, (op, op_arg) in enumerate(code_list):
            # This bit of code is injected between the object on TOS
            # and its pending attribute access. The TOS obj is duped,
            # the rotated above the binder code. The attr is loaded,
            # and the binder is called with the object and attr. The
            # return value of the binder is discarded. This leaves the
            # original TOS and pending attribute access to continue on
            # as normal
            if op == LOAD_ATTR:
                code = [
                    (DUP_TOP, None),
                    (LOAD_CONST, code_binder),
                    (ROT_TWO, None),
                    (LOAD_CONST, op_arg),
                    (CALL_FUNCTION, 0x0002),
                    (POP_TOP, None),
                ]
                insertion_code.append((idx, code))

        return insertion_code

    @abstractmethod
    def monitor_attribute(self, obj, attr):
        """ Hooks up any necessary monitors for the given object and 
        attribute. 

        Parameters
        ----------
        obj : object
            The object on which the attribute access is being performed.
        
        attr : string
            The name of the attribute which is being accessed on the
            object.
        
        """
        raise NotImplementedError

    
#------------------------------------------------------------------------------
# Abstract Call Monitor
#------------------------------------------------------------------------------
class AbstractCallMonitor(AbstractMonitor):
    """ An abstract monitor which monitors the expression evaluation for 
    function calls and calls a method with the object and arguments which
    are being called.
    
    """
    def get_insertion_code(self, code_list):
        """ Generates the byteplay code operations to be inserted into 
        the expression code object to monitor function calls. When an
        attribute access occurs, the 'monitor_function' method will be 
        called with the object, args, and kwargs.

        """
        # A weak closure which is injected into the stack to call the
        # monitor method on function call.
        def code_binder(func_obj, arg_tuple, arg_spec, selfref=ref(self)):
            this = selfref()
            if this is not None:
                nargs = arg_spec & 0xFF
                args = arg_tuple[:nargs]
                kwargs = dict(zip(arg_tuple[nargs::2], arg_tuple[nargs+1::2]))
                this.monitor_function(func_obj, args, kwargs)
            # The UNPACK_SEQUENCE op_codes which will unpack these 
            # return values will unpack things onto the stack in the
            # reverse of how they are provided. So, we pre-reverse them
            # so they come out in the right order.
            return (tuple(reversed(arg_tuple)), func_obj)
                
        # The list of code segments that will be inserted into the
        # new bytecode for the expression.
        insertion_code = []

        for idx, (op, op_arg) in enumerate(code_list):
            # This bit of code is injected just before a function call
            # is performed. The arguments on the stack are packed into
            # tuple. The code binder is then pushed onto the stack and
            # rotated under the func_obj and arg_tuple. The arg spec
            # is then loaded and the code binder is invoked. The return 
            # value of the code binder is the original func_obj and 
            # arg_tuple. This return tuple is unpacked, and then the 
            # arg_tuple is unpacked and the function call proceeds as 
            # normal.
            if op == CALL_FUNCTION:
                # This computes the number of objects on the stack 
                # between TOS and the object being called. Only the
                # last 16bits of the op_arg are signifcant. The lowest
                # 8 are the number of positional args on the stack,
                # the upper 8 is the number of kwargs. For kwargs, the
                # number of items on the stack is twice this number 
                # since the values on the stack alternate name, value.
                n_stack_args = (op_arg & 0xFF) + 2 * ((op_arg >> 8) & 0xFF)
                code = [
                    (BUILD_TUPLE, n_stack_args),
                    (LOAD_CONST, code_binder),
                    (ROT_THREE, None),
                    (LOAD_CONST, op_arg),
                    (CALL_FUNCTION, 0x0003),
                    (UNPACK_SEQUENCE, 2),
                    (UNPACK_SEQUENCE, n_stack_args),
                ]
                insertion_code.append((idx, code))
            
            # TODO - CALL_FUNCTION_VAR, CALL_FUNCTION_KW, CALL_FUNCTION_VAR_KW

        return insertion_code
    
    @abstractmethod
    def monitor_function(self, func_obj, args, kwargs):
        """ Hooks up any necessary monitors for the given function object
        and the args and kwargs with which it is being called.

        Parameters
        ----------
        func_obj : callable object
            The function-like object which is being called.
        
        args : tuple
            The arguments being passed to the function.

        kwargs : dict
            The keyword arguments being passed to the function.
        
        """
        raise NotImplementedError


#------------------------------------------------------------------------------
# Trait Notification Handler
#------------------------------------------------------------------------------
class _TraitNotificationHandler(object):
    """ A thin class which makes it easier to manage the lifetime of 
    a trait notifier.

    """
    def __init__(self, parent, obj, attr):
        """ Initialize a TraitNotificationHandler.

        Parameters
        ----------
        parent : AbstractMonitor
            The AbstractMonitor instance which is the parent of this
            handler. Only a weak reference to the parent is kept.

        obj : HasTraits
            The HasTraits instance on which we are attaching a listener.
        
        attr : string
            The trait attribute on the object to which we should listen.

        """
        self._parent_ref = ref(parent)
        obj.on_trait_change(self.notify, attr)

    def notify(self):
        """ The trait change callback which will emit the expression
        changed signal on the parent.

        """
        parent = self._parent_ref()
        if parent is not None:
            parent.expression_changed()


#------------------------------------------------------------------------------
# Trait Handler Mixin
#------------------------------------------------------------------------------
class TraitHandlerMixin(object):
    """ A mixin class which adds the common binding code for the trait
    monitor classes below.

    """
    def __init__(self, *args, **kwargs):
        super(TraitHandlerMixin, self).__init__(*args, **kwargs)
        # A dictionary which holds the notification handlers. A key
        # in the dictionary is an (obj_id, attr) tuple and the value
        # is the notification handler for that pair. The object id 
        # is used to avoid ref cycles and potential hashing issues.
        self._handlers = {}

    def reset(self):
        """ Clears the existing handlers which results in the notifiers
        being destroyed.

        """
        self._handlers.clear()

    def do_binding(self, obj, attr):
        """ Hooks up a notifier to the object attribute pair if the 
        object is a HasTraits instance and the attribute refers to 
        a valid trait on the object.

        """
        # Don't hook up multiple identifiers to the same object/attr
        # pair. Use the id of the object to prevent an accidental ref
        # cycle to the object.
        handlers = self._handlers
        key = (id(obj), attr)
        if key in handlers:
            return
        
        if isinstance(obj, HasTraits):
            # Only hook up a notifier if the attribute access refers to
            # a proper trait. We check for Disallow trait types since 
            # those can be returned by instances of HasStrictTraits
            trait = obj.trait(attr)
            if trait is not None and trait.trait_type is not Disallow:
                handler = _TraitNotificationHandler(self, obj, attr)
                handlers[key] = handler


#------------------------------------------------------------------------------
# Trait Attribute Monitor
#------------------------------------------------------------------------------
class TraitAttributeMonitor(TraitHandlerMixin, AbstractAttributeMonitor):
    """ An AbstractAttributeMonitor implementation which binds listeners
    to trait attributes on HasTraits classes.

    """
    def monitor_attribute(self, obj, attr):
        """ Hooks up any necessary trait change notifiers for the given
        object and attribute. 

        Parameters
        ----------
        obj : object
            The object on which the attribute access is being performed.
        
        attr : string
            The name of the attribute which is being accessed on the
            object.
        
        """
        self.do_binding(obj, attr)


#------------------------------------------------------------------------------
# Trait Getattr Monitor
#------------------------------------------------------------------------------
class TraitGetattrMonitor(TraitHandlerMixin, AbstractCallMonitor):
    """ An AbstractCallMonitor implementation which binds listeners to
    trait attributes on HasTraits classes which are accessed through a
    call to the builtin getattr.

    """
    def monitor_function(self, func_obj, args, kwargs):
        """ Hooks up any necessary trait change notifiers for the given
        call to getattr

        Parameters
        ----------
        func_obj : callable object
            The function-like object which is being called. If this is
            not the builtin getattr, or if the call spec is invalid for
            getattr, this method is a no-op.
        
        args : tuple
            The arguments being passed to the function.

        kwargs : dict
            The keyword arguments being passed to the function.
        
        """
        n_args = len(args)
        if func_obj is not getattr or n_args < 2 or n_args > 3 or kwargs:
            return
        
        obj, attr = args[0], args[1]
        if not isinstance(attr, basestring):
            return
        
        self.do_binding(obj, attr)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .wx_test_assistant import WXTestAssistant, skip_nonwindows
from .. import progress_bar


@skip_nonwindows
class TestWXProgressBar(WXTestAssistant, progress_bar.TestProgressBar):
    """ WXProgressBar tests.

    .. note:: The wx test uses the extended signature because the range and
        value of the progress bar is related to the minimum attribute.

    """

    def get_value(self, component, widget):
        """  Get the toolkits widget's active value.

        """
        return widget.GetValue() + component.minimum

    def get_minimum(self, component, widget):
        """  Get the toolkits widget's maximum value attribute.

        .. note:: The minimum value is maintained only in the component

        """
        return component.minimum

    def get_maximum(self, component, widget):
        """ Get the toolkits widget's minimum value attribute.

        """
        return widget.GetRange() - component.minimum



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod

from .base_component import BaseComponent
from .toolkit import Toolkit


class EnamlFactory(object):
    """ An abstract base class that defines the interface required
    to instantiate and Enaml declarative component. This class handles
    registering a factory as a virtual base class of the component
    instance and makes sure that a component is associated with a
    proper toolkit.

    """
    __metaclass__ = ABCMeta

    def __call__(self, **kwargs):
        """ Invokes the underlying Enaml build function and applies the
        given keyword arguments as attributes to the result. 

        This a convienence method of the factory which makes is callable
        from Python code. This method will not be called by the Enaml
        runtime.

        Paramters
        ---------
        **kwargs
            The initial attribute values to apply to the component after 
            it as been instantiated.

        Returns
        -------
        result : BaseComponent
            The BaseComponent instance that was created.

        """
        component = self.__enaml_call__()
        component.trait_set(**kwargs)
        return component

    def __enaml_call__(self, identifiers=None, toolkit=None):
        """ Invokes the underlying Enaml build function, creating the 
        locals and toolkit if necessary.

        Parameters
        ----------
        identifiers : dict or None, optional
            The dict of identifiers to use when binding expressions
            on the component. If None, a new dict will be created.
        
        toolkit : Toolkit or None, optional
            The Toolkit to use when building the object tree. If None,
            the currently active toolkit will be used.

        Returns
        -------
        result : BaseComponent
            The BaseComponent instance that was created.

        """
        if identifiers is None:
            identifiers = {}
        if toolkit is None:
            toolkit = Toolkit.active_toolkit()
        component = self.build(identifiers, toolkit)
        component.toolkit = toolkit
        component._bases.insert(0, self)
        return component

    def __instancecheck__(self, instance):
        """ Overrides isinstance(obj, ctor) for instances of this factory
        class. This allows one check if a component instance was created 
        by a given factory instance using standard Python idioms.

        """
        if isinstance(instance, BaseComponent):
            return self in instance._bases
        return False

    @abstractmethod
    def build(self, identifiers, toolkit):
        """ An abstract method which must be implemented by subclasses
        to build an return an instance of BaseComponent.

        Parameters
        ----------
        identifiers : dict
            The dict of identifiers to use when binding expressions
            on the component.
        
        toolkit : Toolkit
            The Toolkit to use when building the object tree.

        Returns
        -------
        result : BaseComponent
            The BaseComponent instance that was created.

        """
        raise NotImplementedError


#------------------------------------------------------------------------------
# Enaml Declaration
#------------------------------------------------------------------------------
class EnamlDeclaration(EnamlFactory):
    """ An EnamlFactory which exposes a compiled Enaml declaration
    function with an interface that is easy to use from Python.

    """
    def __init__(self, func):
        self.__func__ = func
        self.__doc__ = func.__doc__
        self.__name__ = func.__name__
        self.__module__ = func.__module__

    def __repr__(self):
        return '%s.%s' % (self.__module__, self.__name__)

    def build(self, identifiers, toolkit):
        """ An abstractmethod implementation that returns the Enaml
        component generated by the internal compiled Enaml function.

        """
        return self.__func__(identifiers, toolkit)




# -*- coding: UTF-8 -*-
#------------------------------------------------------------------------------
#  file: class_doc.py
#  License: LICENSE.TXT
#  Author: Ioannis Tziakos
#
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from base_doc import BaseDoc
from line_functions import get_indent, replace_at, add_indent, fix_backspace
from fields import AttributeField


class EnamlDeclDoc(BaseDoc):
    """Docstring refactoring for classes"""

    def __init__(self, lines, headers=None, verbose=False):

        if headers is None:
            headers = {'Input Attributes': 'attributes',
                       'Synchronized Attributes': 'attributes',
                       'Output Attributes': 'attributes',
                       'Public Attributes': 'attributes'}

        super(EnamlDeclDoc, self).__init__(lines, headers, verbose)
        return

    def _refactor_attributes(self, header):
        """Refactor the attributes section to sphinx friendly format"""

        if self.verbose:
            print '{0} Section'.format(header)

        index = self.index
        self.remove_lines(index, 2)
        indent = get_indent(self.peek())
        fields = self.extract_fields(indent, AttributeField)
        header = fix_backspace(header)
        lines = [indent + ':{0}:'.format(header), '']
        for field in fields:
            lines += field.to_rst(len(indent) + 4)
        self.insert_lines(lines[:-1], index)
        self.index += len(lines)
        return

    def _refactor_notes(self, header):
        """Refactor the note section to use the rst ``.. note`` directive.

        """
        descriptions = []
        index = self.index
        self.remove_lines(index, 2)
        indent = get_indent(self.peek())
        paragraph = self.get_next_paragraph()
        descriptions.append(indent + '.. note::')
        descriptions += add_indent(paragraph)
        self.insert_lines(descriptions, index)
        self.index += len(descriptions)
        return descriptions



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .enaml_test_case import EnamlTestCase, required_method


class TestHtml(EnamlTestCase):
    """ Logic for testing Html widgets.

    The toolkits return HTML as plain text, so we do not compare formatting.
    Tooklit testcases need to provide the following methods

    Abstract Methods
    ----------------
    get_source(self, widget)
        Get the source of an Html widget.

    """
    text = 'That is a bold claim.'

    def setUp(self):
        """ Set up before the Html tests.

        """
        enaml_source = """
enamldef MainView(MainWindow):
    Html:
        name = 'html'
        source = '<b>{0}</b>'
""".format(self.text)

        self.view = self.parse_and_create(enaml_source)
        self.component = self.component_by_name(self.view, 'html')
        self.widget = self.component.toolkit_widget

    def test_initial_source(self):
        """ Test the initial source of an Html widget.

        """
        widget_source = self.get_source(self.widget)
        self.assertEqual(widget_source, self.text)

    def test_source_changed(self):
        """ Change the source of an Html widget.

        """
        new_text = 'Underlined'
        new_source = '<u>{0}</u>'.format(new_text)
        self.component.source = new_source
        widget_source = self.get_source(self.widget)
        self.assertEqual(widget_source, new_text)

    #--------------------------------------------------------------------------
    # Abstract methods
    #--------------------------------------------------------------------------
    @required_method
    def get_source(self, widget):
        """ Get the source of an Html widget.

        """
        pass




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import HasTraits, BaseTuple

from .color import ColorTrait
from .font import FontTrait
from .layout import AlignmentTrait, MarginsTrait, SizeTrait, SizePolicyTrait, SpacingTrait, StretchTrait


class StyleClassTrait(BaseTuple):

    def validate(self, obj, name, value):
        if isinstance(value, basestring):
            res = tuple(val.strip() for val in value.split())
        elif isinstance(value, (list, tuple)):
            if all(isinstance(val, basestring) for val in value):
                res = tuple(val.strip() for val in value)
            else:
                res = ()
        else:
            res = ()
        return res


class StyleNode(HasTraits):

    #: The alignment of a widget in its parent layout
    alignment = AlignmentTrait

    #: The foreground color of a widget. This is usually text color.
    fgcolor = ColorTrait

    #: The background color of a widget.
    bgcolor = ColorTrait

    #: The font to use for the widget.
    font = FontTrait

    #: The margins to place around a layout.
    margins = MarginsTrait

    #: The suggested (w, h) size of the widget.
    size = SizeTrait

    #: How the widget should be resized when placed in a layout.
    size_policy = SizePolicyTrait

    #: The interelement spacing in a layout
    spacing = SpacingTrait

    #: The stretch factor to use when adding the widget to a layout
    stretch = StretchTrait

    #: The string name of the style class to use for this node.
    style_class = StyleClassTrait




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx

from .wx_widget_component import EVT_MIN_SIZE
from .wx_window import WXWindow

from ...components.main_window import AbstractTkMainWindow


class WXMainWindowSizer(wx.PySizer):
    """ A custom wx Sizer for use in the WXMainWindow. This sizer expands
    its child to fit the allowable space, but respects the minimum size
    of the child.

    There can only be one widget in this sizer at a time and it should
    be added via the .Add(...) method. Old items will be removed 
    automatically (but not destroyed).

    """
    def __init__(self, *args, **kwargs):
        super(WXMainWindowSizer, self).__init__(*args, **kwargs)
        self._widget = None
        self._min_size = None

    def OnWidgetMinSize(self, event):
        """ An event handler which is called when the min size is set 
        on the central widget. It updates the internal cached min size
        and updates the frame sizing.

        """
        self._min_size = self._widget.GetMinSize()
        self.UpdateFrameSizing()
        event.Skip()

    def Add(self, widget):
        """ Adds the given widget to the sizer, removing the old widget
        if present. The old widget is not destroyed.

        """
        self.Clear(deleteWindows=False)
        old_widget = self._widget
        if old_widget is not None:
            old_widget.Unbind(EVT_MIN_SIZE, handler=self.OnWidgetMinSize)
        widget.Bind(EVT_MIN_SIZE, self.OnWidgetMinSize)
        self._widget = widget
        self._min_size = widget.GetMinSize()
        res = super(WXMainWindowSizer, self).Add(widget)
        self.UpdateFrameSizing()
        return res

    def CalcMin(self):
        """ Returns the computed minimum size for the for the frame.

        """
        return self._min_size
    
    def RecalcSizes(self):
        """ Resizes the child to fit the available space of the window.

        """
        widget = self._widget
        if widget:
            widget.SetSize(self.GetSize())
                
    def UpdateFrameSizing(self):
        """ Sets the minimum size allowed of the frame and resizes it 
        if necessary. This is called whenever the minimum size of the
        central widget changes.

        """
        frame = self.GetContainingWindow()
        min_width, min_height = self._min_size
        frame_width, frame_height = frame.GetSizeTuple()
        client_width, client_height = frame.GetClientSizeTuple()
        delta_width = frame_width - client_width
        delta_height = frame_height - client_height
        new_min_width = min_width + delta_width
        new_min_height = min_height + delta_height
        frame.SetMinSize((new_min_width, new_min_height))
        if new_min_width > frame_width or new_min_height > frame_height:
            resize_width = max(new_min_width, frame_width)
            resize_height = max(new_min_height, frame_height)
            frame.SetSize((resize_width, resize_height))


class WXMainWindow(WXWindow, AbstractTkMainWindow):
    """ A Wx implementation of a MainWindow.

    """
    #--------------------------------------------------------------------------
    # Setup Methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying wx.Frame with an appropriate main
        window sizer.

        """
        self.widget = wx.Frame(parent)
        self.widget.SetSizer(WXMainWindowSizer())

    def initialize(self):
        """ Intializes the attributes on the wx.Frame.

        """
        super(WXMainWindow, self).initialize()
        shell = self.shell_obj
        self.set_menu_bar(shell.menu_bar)

    def bind(self):
        """ Bind the signal handlers for the wx.Frame.

        """
        super(WXMainWindow, self).bind()
        self.widget.Bind(wx.EVT_CLOSE, self._on_close)
    
    #--------------------------------------------------------------------------
    # Change Handlers
    #--------------------------------------------------------------------------
    def shell_menu_bar_changed(self, menu_bar):
        """ Update the menu bar of the window with the new value from
        the shell object.

        """
        self.set_menu_bar(menu_bar)

    #--------------------------------------------------------------------------
    # Event Handlers
    #--------------------------------------------------------------------------
    def _on_close(self, event):
        """ Emits the closed event on the shell object when the main 
        window is closed.

        """
        event.Skip()
        self.shell_obj.closed()
    
    #--------------------------------------------------------------------------
    # Widget Update Methods
    #--------------------------------------------------------------------------
    def set_central_widget(self, central_widget):
        """ Sets the central widget in the window with the given value.

        """
        # It's possible for the central widget component to be None.
        # This must be allowed since the central widget may be generated
        # by an Include component, in which case it will not exist 
        # during initialization. However, we must have a central widget
        # for the Dialog, and so we just fill it with a dummy widget.
        central_widget = self.shell_obj.central_widget
        if central_widget is None:
            child_widget = wx.Panel(self.widget)
        else:
            child_widget = central_widget.toolkit_widget
        self.widget.GetSizer().Add(child_widget)

    def set_menu_bar(self, menu_bar):
        """ Updates the menu bar in the main window with the given Enaml
        MenuBar instance.

        """
        if menu_bar is not None:
            self.widget.SetMenuBar(menu_bar.toolkit_widget)
        else:
            self.widget.SetMenuBar(None)

    def set_visible(self, visible):
        """ Overridden from the parent class to raise the window to
        the front if it should be shown.

        """
        widget = self.widget
        if visible:
            widget.Show(True)
            widget.Raise()
        else:
            widget.Show(False)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt import QtGui
from .qt_control import QtControl

from ...components.label import AbstractTkLabel


class QtLabel(QtControl, AbstractTkLabel):
    """ A Qt implementation of Label.

    """
    #: The internal cached size hint which is used to determine whether
    #: of not a size hint updated event should be emitted when the text
    #: in the label changes
    _cached_size_hint = None

    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying QLabel control.

        """
        self.widget = QtGui.QLabel(parent)

    def initialize(self):
        """ Initializes the attributes on the underlying control.

        """
        super(QtLabel, self).initialize()
        shell = self.shell_obj
        self.set_text(shell.text)
        self.set_word_wrap(shell.word_wrap)
        
    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_text_changed(self, text):
        """ The change handler for the 'text' attribute.

        """
        self.set_text(text)

    def shell_word_wrap_changed(self, word_wrap):
        """ The change handler for the 'word_wrap' attribute.

        """
        self.set_word_wrap(word_wrap)

    def set_text(self, text):
        """ Sets the label on the underlying control.

        """
        self.widget.setText(text)

        # Emit a size hint updated event if the size hint has actually
        # changed. This is an optimization so that a constraints update
        # only occurs when the size hint has actually changed. This 
        # logic must be implemented here so that the label has been
        # updated before the new size hint is computed. Placing this
        # logic on the shell object would not guarantee that the label
        # has been updated at the time the change handler is called.
        cached = self._cached_size_hint
        hint = self._cached_size_hint = self.size_hint()
        if cached != hint:
            self.shell_obj.size_hint_updated()

    def set_word_wrap(self, wrap):
        """ Sets the word wrapping on the underlying widget.

        """
        self.widget.setWordWrap(wrap)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx

from .wx_control import WXControl

from ...components.combo_box import AbstractTkComboBox


class WXComboBox(WXControl, AbstractTkComboBox):
    """ A wxPython implementation of ComboBox.

    Use a combo box to select a single item from a collection of items.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates a wx.ComboBox.

        """
        self.widget = wx.ComboBox(parent, style=wx.CB_READONLY)

    def initialize(self):
        """ Intializes the widget with the attributes of this instance.

        """
        super(WXComboBox, self).initialize()
        shell = self.shell_obj
        self.set_items(shell.labels)
        self.set_selection(shell.index)

    def bind(self):
        """ Binds the event handlers for the combo box.

        """
        super(WXComboBox, self).bind()
        self.widget.Bind(wx.EVT_COMBOBOX, self.on_selected)

    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_index_changed(self, index):
        """ The change handler for the '_index' attribute on the enaml
        shell.

        """
        self.set_selection(index)

    def shell_labels_changed(self, labels):
        """ The change handler for the 'labels' attribute on the shell
        object.

        """
        self.set_items(labels)

    def on_selected(self, event):
        """ The event handler for a combo box selection event.

        """
        shell = self.shell_obj
        curr_index = self.widget.GetCurrentSelection()
        shell.index = curr_index

        # Only fire the selected event if we have a valid selection
        if curr_index != -1:
            shell.selected(shell.value)

        event.Skip()

    def set_items(self, str_items):
        """ Sets the items in the combo box.

        """
        # wx does not emit events on SetItems or SetSelection, so we
        # do not need any feedback guards here.
        self.widget.SetItems(str_items)
        self.widget.SetSelection(self.shell_obj.index)

    def set_selection(self, index):
        """ Set the selected value in the toolkit widget.

        """
        # wx does not emit events on SetSelection, so we do 
        # not need any feedback guards here.
        self.widget.SetSelection(index) # deselects if -1




# -*- coding: UTF-8 -*-
#------------------------------------------------------------------------------
#  file: function_doc.py
#  License: LICENSE.TXT
#  Author: Ioannis Tziakos
#
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from base_doc import BaseDoc
from line_functions import get_indent, add_indent
from fields import ArgumentField, ListItemWithTypeField, ListItemField


class FunctionDoc(BaseDoc):
    """Docstring refactoring for functions"""

    def __init__(self, lines, headers=None, verbose=False):

        if headers is None:
            headers = {'Returns': 'returns', 'Arguments': 'arguments',
                       'Parameters': 'arguments', 'Raises': 'raises',
                       'Yields': 'returns', 'Notes':'notes'}

        super(FunctionDoc, self).__init__(lines, headers, verbose)
        return

    def _refactor_returns(self, header):
        """Refactor the return section to sphinx friendly format.

        """
        index = self.index
        self.remove_lines(index, 2)
        indent = get_indent(self.peek())
        fields = self.extract_fields(indent, field_type=ListItemWithTypeField)
        lines = [indent + ':returns:']
        prefix = '' if len(fields) == 1 else '- '
        for field in fields:
            lines += field.to_rst(len(indent) + 4, prefix)
        self.insert_lines(lines, index)
        self.index += len(lines)
        return

    def _refactor_raises(self, header):
        """Refactor the raises section to sphinx friendly format"""
        index = self.index
        self.remove_lines(index, 2)
        indent = get_indent(self.peek())
        fields = self.extract_fields(indent, field_type=ListItemField)
        lines = [indent + ':raises:']
        prefix = '' if len(fields) == 1 else '- '
        for field in fields:
            lines += field.to_rst(len(indent) + 4, prefix)
        self.insert_lines(lines, index)
        self.index += len(lines)
        return

    def _refactor_arguments(self, header):
        """Refactor the argument section to sphinx friendly format
        """
        index = self.index
        self.remove_lines(index, 2)
        indent = get_indent(self.peek())
        fields = self.extract_fields(indent, field_type=ArgumentField)
        lines = []
        for field in fields:
            lines += field.to_rst(len(indent))
        self.insert_lines(lines, index)
        self.index += len(lines)
        return

    def _refactor_notes(self, header):
        """Refactor the argument section to sphinx friendly format.

        """
        if self.verbose:
            print 'Refactoring Notes'

        descriptions = []
        index = self.index
        self.remove_lines(index, 2)
        indent = get_indent(self.peek())
        paragraph = self.get_next_paragraph()
        descriptions.append(indent + '.. note::')
        descriptions += add_indent(paragraph)
        self.insert_lines(descriptions, index)
        self.index += len(descriptions)
        return descriptions




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import abstractmethod

from traits.api import Bool, Str, Instance, Property

from .control import Control, AbstractTkControl

from ..core.trait_types import CoercingInstance, EnamlEvent
from ..layout.geometry import Size
from ..noncomponents.abstract_icon import AbstractTkIcon


class AbstractTkPushButton(AbstractTkControl):
    """ The abstract toolkit interface for a PushButton.

    """
    @abstractmethod
    def shell_text_changed(self):
        """ The change handler for the 'text' attribute on the shell 
        component.

        """
        raise NotImplementedError

    @abstractmethod
    def shell_icon_changed(self, icon):
        """ The change handler for the 'icon' attribute on the shell
        component.

        """
        raise NotImplementedError

    @abstractmethod
    def shell_icon_size_changed(self, icon_size):
        """ The change handler for the 'icon_size' attribute on the shell
        component.

        """
        raise NotImplementedError


class PushButton(Control):
    """ A push button widget.

    """
    #: A read only property which indicates whether or not the button 
    #: is currently pressed.
    down = Property(Bool, depends_on='_down')

    #: The text to use as the button's label.
    text = Str

    #: The an icon to display in the button.
    icon = Instance(AbstractTkIcon)

    # The size of the icon
    icon_size = CoercingInstance(Size)
    
    #: Fired when the button is pressed and released.
    clicked = EnamlEvent
    
    #: Fired when the button is pressed.
    pressed = EnamlEvent

    #: Fired when the button is released.
    released = EnamlEvent
    
    #: How strongly a component hugs it's contents' width. PushButtons 
    #: hug their contents' width weakly.
    hug_width = 'weak'

    #: An internal attribute that is used by the implementation object
    #: to set the value of down.
    _down = Bool

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkPushButton)

    def _get_down(self):
        """ The property getter for the 'down' attribute.

        """
        return self._down




#------------------------------------------------------------------------------
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import os

from .qt.QtGui import QFileDialog
from .qt_dialog import QtDialog

from ...components.directory_dialog import AbstractTkDirectoryDialog
from ...guard import guard


class QtDirectoryDialog(QtDialog, AbstractTkDirectoryDialog):
    """ A Qt4 implementation of a DirectoryDialog.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying QFileDialog control.

        """
        self.widget = QFileDialog(parent)

    def initialize(self):
        """ Initializes the attributes of the underlying QFileDialog.

        """
        super(QtDirectoryDialog, self).initialize()
        shell = self.shell_obj
        self.set_directory(shell.directory)
        self.widget.setAcceptMode(QFileDialog.AcceptOpen)
        self.widget.setFileMode(QFileDialog.DirectoryOnly)
        self.widget.setViewMode(QFileDialog.Detail)

    def bind(self):
        """ Binds the signal handlers for the file dialog.

        """
        super(QtDirectoryDialog, self).bind()
        widget = self.widget
        widget.filesSelected.connect(self.on_files_selected)

    #--------------------------------------------------------------------------
    # Shell Object Change Handlers
    #--------------------------------------------------------------------------
    def shell_directory_changed(self, directory):
        """ Update the dialog with the new working directory.

        """
        if not guard.guarded(self, 'setting_directory'):
            self.set_directory(directory)
                
    #--------------------------------------------------------------------------
    # Event Handlers 
    #--------------------------------------------------------------------------
    def on_files_selected(self, files):
        """ The signal handler for the dialog's `filesSelected` signal.

        """
        shell = self.shell_obj
        first_file = files[0] if files else u''
        with guard(self, 'setting_directory'):
            shell.directory = first_file

    #--------------------------------------------------------------------------
    # Widget Update Methods
    #--------------------------------------------------------------------------
    def set_directory(self, directory):
        """ Sets the current directory of the dialog.

        """
        self.widget.setDirectory(directory)

    def set_central_widget(self, widget):
        """ Overridden parent class method which makes central widget
        operations a no-op, since a file dialog has no children.

        """
        pass




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt.QtCore import Signal
from .qt.QtGui import QLineEdit
from .qt_control import QtControl
from .qt_enaml_validator import QtEnamlValidator

from ...components.field import AbstractTkField
from ...guard import guard


_PASSWORD_MODES = {
    'normal': QLineEdit.Normal,
    'password': QLineEdit.Password,
    'silent': QLineEdit.NoEcho,
}


class QtEnamlLineEdit(QLineEdit):
    """ A QLineEdit subclass which converts a lost focus event into 
    a lost focus signal.

    """
    lostFocus = Signal()

    def focusOutEvent(self, event):
        self.lostFocus.emit()
        return super(QtEnamlLineEdit, self).focusOutEvent(event)


class QtField(QtControl, AbstractTkField):
    """ A Qt implementation of a Field which uses a QLineEdit to provide
    a single line of editable text.

    """
    #--------------------------------------------------------------------------
    # SetupMethods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying QLineEdit.

        """
        self.widget = QtEnamlLineEdit(parent)

    def initialize(self):
        """ Initializes the attributes of the Qt widget.

        """
        super(QtField, self).initialize()
        shell = self.shell_obj
        self.set_validator(shell.validator)
        self.set_read_only(shell.read_only)
        self.set_placeholder_text(shell.placeholder_text)
        self.set_text(shell.validator.format(shell.value))
        self.set_cursor_position(shell.cursor_position)
        self.set_password_mode(shell.password_mode)
        self.set_max_length(shell.max_length)
        
    def bind(self):
        """ Binds the event handlers for the QLineEdit.

        """
        super(QtField, self).bind()
        widget = self.widget
        widget.textEdited.connect(self.on_text_edited)
        widget.returnPressed.connect(self.on_return_pressed)
        widget.lostFocus.connect(self.on_lost_focus)
        widget.selectionChanged.connect(self.on_selection_changed)
        widget.cursorPositionChanged.connect(self.on_cursor_changed)

    #--------------------------------------------------------------------------
    # Shell Object Change Handlers
    #-------------------------------------------------------------------------- 
    def shell_validator_changed(self, validator):
        """ The change handler for the 'validator' attribute on the 
        shell object.
        
        """
        self.set_validator(validator)
           
    def shell_max_length_changed(self, max_length):
        """ The change handler for the 'max_length' attribute on the 
        shell object.

        """
        self.set_max_length(max_length)

    def shell_read_only_changed(self, read_only):
        """ The change handler for the 'read_only' attribute on the
        shell object.

        """
        self.set_read_only(read_only)

    def shell_cursor_position_changed(self, cursor_position):
        """ The change handler for the 'cursor_position' attribute on 
        the shell object.

        """
        if not guard.guarded(self, 'updating_cursor'):
            self.set_cursor_position(cursor_position)

    def shell_placeholder_text_changed(self, placeholder_text):
        """ The change handler for the 'placeholder_text' attribute on
        the shell object.

        """
        self.set_placeholder_text(placeholder_text)

    def shell_password_mode_changed(self, mode):
        """ The change handler for the 'password_mode' attribute on the 
        shell object.
        
        """
        self.set_password_mode(mode)

    #--------------------------------------------------------------------------
    # Manipulation Methods 
    #--------------------------------------------------------------------------
    def set_selection(self, start, end):
        """ Sets the selection to the bounds of start and end.

        If the indices are invalid, no selection will be made, and any
        current selection will be cleared.

        Arguments
        ---------
        start : Int
            The start selection index, zero based.

        end : Int
            The end selection index, zero based.

        """
        self.widget.setSelection(start, end - start)

    def select_all(self):
        """ Select all the text in the field.

        If there is no text in the field, the selection will be empty.

        """
        self.widget.selectAll()

    def deselect(self):
        """ Deselect any selected text.

        """
        self.widget.deselect()

    def clear(self):
        """ Clear the field of all text.

        """
        self.widget.clear()

    def backspace(self):
        """ Simple backspace functionality.

        If no text is selected, deletes the character to the left of 
        the cursor. Otherwise, it deletes the selected text.

        """
        self.widget.backspace()

    def delete(self):
        """ Simple delete functionality.

        If no text is selected, deletes the character to the right of 
        the cursor. Otherwise, it deletes the selected text.

        """
        self.widget.del_()

    def end(self, mark=False):
        """ Moves the cursor to the end of the line.

        Arguments
        ---------
        mark : bool, optional
            If True, select the text from the current position to the 
            end of the field. Defaults to False.

        """
        widget = self.widget
        if mark:
            start = widget.cursorPosition()
            end = len(widget.text())
            widget.setSelection(start, end)
        else:
            widget.end(mark)

    def home(self, mark=False):
        """ Moves the cursor to the beginning of the line.

        Arguments
        ---------
        mark : bool, optional
            If True, select the text from the current position to the 
            beginning of the field. Defaults to False.

        """
        widget = self.widget
        if mark:
            start = 0
            end = widget.cursorPosition()
            widget.setSelection(start, end)
        else:
            widget.home(mark)

    def cut(self):
        """ Cuts the selected text from the field.

        Copies the selected text to the clipboard then deletes the 
        selected text from the field.

        """
        self.widget.cut()

    def copy(self):
        """ Copies the selected text to the clipboard.

        """
        self.widget.copy()

    def paste(self):
        """ Paste the contents of the clipboard into the field.

        Inserts the contents of the clipboard into the field at the 
        current cursor position, replacing any selected text.

        """
        self.widget.paste()

    def insert(self, text):
        """ Insert the text into the field.

        Inserts the given text into the field at the current cursor 
        position, replacing any selected text.

        Arguments
        ---------
        text : str
            The text to insert into the field.

        """
        self.widget.insert(text)

    def undo(self):
        """ Undoes the last operation.

        """
        self.widget.undo()

    def redo(self):
        """ Redoes the last operation.

        """
        self.widget.redo()

    #--------------------------------------------------------------------------
    # Signal Handlers 
    #--------------------------------------------------------------------------
    def on_text_edited(self):
        """ The signal handler for the 'textEdited' signal. This is
        called when the user edits the text via the ui.

        """
        self.shell_obj._field_text_edited()

    def on_return_pressed(self):
        """ The signal handler for the returnPressed signal.

        """
        self.shell_obj._field_return_pressed()

    def on_lost_focus(self):
        """ The signal handler for the lostFocus signal.

        """
        self.shell_obj._field_lost_focus()

    def on_selection_changed(self):
        """ The signal handler for the selectionChanged signal.

        """
        with guard(self, 'updating_selection'):
            self.shell_obj.selected_text = self.widget.selectedText()

    def on_cursor_changed(self):
        """ The event handler for a cursor change event.

        """
        with guard(self, 'updating_cursor'):
            self.shell_obj.cursor_position = self.widget.cursorPosition()

    #--------------------------------------------------------------------------
    # Update methods 
    #--------------------------------------------------------------------------
    def get_text(self):
        """ Returns the current unicode text in the control.

        """
        return self.widget.text()
    
    def set_text(self, text):
        """ Updates the text control with the given unicode text.

        """
        self.widget.setText(text)

    def set_validator(self, validator):
        """ Wraps the given Enaml validator in a custom QValidator 
        instance and applies it to the underlying control.

        """
        qvalidator = QtEnamlValidator(validator)
        self.widget.setValidator(qvalidator)
    
    def set_max_length(self, max_length):
        """ Set the max length of the control to max_length. If the max 
        length is <= 0 or > 32767 then the control will be set to hold 
        32kb of text.

        """
        if (max_length <= 0) or (max_length > 32767):
            max_length = 32767
        self.widget.setMaxLength(max_length)

    def set_read_only(self, read_only):
        """ Sets read only state of the widget.

        """
        self.widget.setReadOnly(read_only)

    def set_placeholder_text(self, placeholder_text):
        """ Sets the placeholder text in the widget.

        """
        self.widget.setPlaceholderText(placeholder_text)

    def set_cursor_position(self, cursor_position):
        """ Sets the cursor position of the widget.

        """
        self.widget.setCursorPosition(cursor_position)

    def set_password_mode(self, password_mode):
        """ Sets the password mode of the wiget.

        """
        self.widget.setEchoMode(_PASSWORD_MODES[password_mode])
        



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx
import wx.grid

from .wx_control import WXControl

from ...components.table_view import AbstractTkTableView
from ...item_models.abstract_item_model import AbstractItemModel, ITEM_IS_EDITABLE


GridCellAttr = wx.grid.GridCellAttr


def wx_color_from_color(color):
    return wx.Color(*color)


# XXX we need to add more handler support for the different modes of
# resetting the grid.
class AbstractItemModelTable(wx.grid.PyGridTableBase):

    def __init__(self, item_model):
        super(AbstractItemModelTable, self).__init__()
        if not isinstance(item_model, AbstractItemModel):
            raise TypeError('Model must be an instance of AbstractItemModel.')
        self._item_model = item_model

        # Cache frequently used bound methods
        self._row_count = item_model.row_count
        self._col_count = item_model.column_count
        self._model_index = item_model.index
        self._model_data = item_model.data
        self._model_background = item_model.background
        self._model_foreground = item_model.foreground
        self._vert_header_data = item_model.vertical_header_data
        self._horiz_header_data = item_model.horizontal_header_data

        self._item_model.model_reset.connect(self._end_model_reset)
        self._item_model.data_changed.connect(self._data_changed)

    def _end_model_reset(self):
        grid = self.GetView()
        grid.SetTable(self)
        grid.Refresh()

    def _data_changed(self):
        grid = self.GetView()
        flag = wx.grid.GRIDTABLE_REQUEST_VIEW_GET_VALUES
        msg = wx.grid.GridTableMessage(self, flag)
        grid.ProcessTableMessage(msg)
        grid.Refresh()

    def GetNumberRows(self):
        return self._row_count()

    def GetNumberCols(self):
        return self._col_count()

    def GetValue(self, row, col):
        index = self._model_index(row, col, None)
        return self._model_data(index)

    def GetAttr(self, row, col, ignored):
        index = self._model_index(row, col, None)

        attr = GridCellAttr()

        brush = self._model_background(index)
        if brush is not None:
            attr.SetBackgroundColour(wx_color_from_color(brush.color))

        brush = self._model_foreground(index)
        if brush is not None:
            attr.SetTextColour(wx_color_from_color(brush.color))
            
        return attr

    def SetValue(self, row, col, val):
        item_model = self._item_model
        index = item_model.index(row, col, None)
        flags = item_model.flags(index)
        if flags & ITEM_IS_EDITABLE:
            item_model.set_data(index, val)

    def GetRowLabelValue(self, row):
        return self._vert_header_data(row)

    def GetColLabelValue(self, col):
        return self._horiz_header_data(col)


class WXTableView(WXControl, AbstractTkTableView):
    """ A wxPython implementation of TableView.

    """
    #: The underlying model.
    model_wrapper = None

    #---------------------------------------------------------------------------
    # ITableViewImpl interface
    #---------------------------------------------------------------------------
    def create(self, parent):
        """ Create the underlying wx.grid.Grid control.

        """
        style = wx.WANTS_CHARS | wx.FULL_REPAINT_ON_RESIZE
        self.widget = wx.grid.Grid(parent, style=style)

    def initialize(self):
        """ Intialize the widget with the attributes of this instance.

        """
        super(WXTableView, self).initialize()
        self.set_table_model(self.shell_obj.item_model)
        self.widget.SetDoubleBuffered(True)

    def bind(self):
        """ Bind the event handlers for the table view.

        """
        super(WXTableView, self).bind()
        widget = self.widget
        widget.Bind(wx.grid.EVT_GRID_SELECT_CELL, self.on_select_cell)
        widget.Bind(wx.grid.EVT_GRID_CELL_LEFT_DCLICK, self.on_left_dclick)

    #---------------------------------------------------------------------------
    # implementation
    #---------------------------------------------------------------------------
    def shell_item_model_changed(self, item_model):
        """ The change handler for the 'item_model' attribute.

        """
        self.set_table_model(item_model)

    def shell_vertical_header_visible_changed(self, visible):
        """ The change handler for the 'vertical_header_visible' 
        attribute of the shell object.

        """
        pass

    def shell_horizontal_header_visible_changed(self, visible):
        """ The change handler for the 'horizontal_header_visible'
        attribute of the shell object.

        """
        pass
        
    def on_select_cell(self, event):
        """ The event handler for the cell selection event.  Not meant
        for public consumption.

        """
        event.Skip()

    def on_left_dclick(self, event):
        """ The event handler for the double-click event.  Not meant
        for public consumption.

        """
        event.Skip()

    def set_table_model(self, model):
        """ Set the table view's model.  Not meant for public
        consumption.

        """
        model_wrapper = AbstractItemModelTable(model)
        self.widget.SetTable(model_wrapper)
        self.widget.Refresh()
        self.model_wrapper = model_wrapper

    #--------------------------------------------------------------------------
    # Overrides
    #--------------------------------------------------------------------------
    def size_hint(self):
        """ A reimplemented parent class method which computes the size 
        hint for the table view.

        """
        # This size hint is taken from QAbstractScrollArea::sizeHint
        # which is the size hint implementation for QTableView.
        # The wx GetBestSize() call returns the size for the entire
        # table, which may be millions of pixels for large tables.
        # It also takes forever and a year to compute.
        return (256, 192)
        



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .expressions import (
    SimpleExpression, NotificationExpression, SubscriptionExpression, 
    DelegationExpression, UpdateExpression,
)
from .inverters import (
    GenericAttributeInverter, GetattrInverter, ImplicitAttrInverter,
)
from .monitors import TraitAttributeMonitor, TraitGetattrMonitor


#------------------------------------------------------------------------------
# Builtin Enaml Operators
#------------------------------------------------------------------------------
#: Operators are passed a number of arguments from the Enaml runtime 
#: engine in order to perform their work. The arguments are, in order:
#:
#:      cmpnt : BaseComponent
#:          The BaseComponent instance which owns the expression which
#:          is being bound.
#:
#:      attr : string
#:          The name of the attribute on the component which is being
#:          bound.
#:
#:      code : types.CodeType
#:          The compiled code object for the Python expression given
#:          given by the user. It is compiled with mode='eval'.
#:
#:      identifiers : dict
#:          The dictionary of identifiers available to the expression.
#:          This dictionary is shared amongs all expressions within
#:          a given lexical scope. It should therefore not be modified
#:          or copied since identifiers may continue to be added to 
#:          this dict as runtime execution continues.
#:
#:      f_globals : dict
#:          The dictionary of globals available to the  expression. 
#:          The same rules about sharing and copying that apply to
#:          the identifiers dict, apply here as well.
#:
#:      toolkit : Toolkit
#:          The toolkit instance that is in scope for the expression.
#:          The same rules about sharing and copying that apply to
#:          the identifiers dict, apply here as well.
#:
#: Operators may do whatever they please with the information provided
#: to them. The default operators in Enaml use this information to 
#: create and bind Enaml expression objects to the component. However,
#: this is not a requirement and developers who are extending enaml
#: are free to get creative with the operators.


def op_simple(cmpnt, attr, code, identifiers, f_globals, toolkit):
    """ The default Enaml operator for '=' expressions. It binds an
    instance of SimpleExpression to the component.

    """
    expr = SimpleExpression(cmpnt, attr, code, identifiers, f_globals, toolkit)
    cmpnt.bind_expression(attr, expr)


def op_notify(cmpnt, attr, code, identifiers, f_globals, toolkit):
    """ The default Enaml operator for '::' expressions. It binds an
    instance of NotificationExpression to the component.

    """
    expr = NotificationExpression(cmpnt, attr, code, identifiers, f_globals, toolkit)
    cmpnt.bind_expression(attr, expr, notify_only=True)


def op_update(cmpnt, attr, code, identifiers, f_globals, toolkit):
    """ The default Enaml operator for '>>' expressions. It binds an
    instance of UpdateExpression to the component.

    """
    inverters = [GenericAttributeInverter, GetattrInverter, ImplicitAttrInverter]
    expr = UpdateExpression(inverters, cmpnt, attr, code, identifiers, f_globals, toolkit)
    cmpnt.bind_expression(attr, expr, notify_only=True)


def op_subscribe(cmpnt, attr, code, identifiers, f_globals, toolkit):
    """ The default Enaml operator for '<<' expressions. It binds an
    instance of SubscriptionExpression to the component using monitors
    which understand traits attribute access via dotted notation and
    the builtin getattr function.

    """
    monitors = [TraitAttributeMonitor, TraitGetattrMonitor]
    expr = SubscriptionExpression(monitors, cmpnt, attr, code, identifiers, f_globals, toolkit)
    cmpnt.bind_expression(attr, expr)


def op_delegate(cmpnt, attr, code, identifiers, f_globals, toolkit):
    """ The default Enaml operator for ':=' expressions. It binds an
    instance of DelegationExpression to the component using monitors
    which understand traits attribute access via dotted notation and
    the builtin getattr function, and inverters which understand the
    dotted attribute access, implicit attribute access, and also the
    builtin getattr function.

    """
    inverters = [GenericAttributeInverter, GetattrInverter, ImplicitAttrInverter]
    monitors = [TraitAttributeMonitor, TraitGetattrMonitor]
    expr = DelegationExpression(inverters, monitors, cmpnt, attr, code, identifiers, f_globals, toolkit)
    cmpnt.bind_expression(attr, expr)


OPERATORS = {
    '__operator_Equal__': op_simple,
    '__operator_LessLess__': op_subscribe,
    '__operator_ColonEqual__': op_delegate,
    '__operator_ColonColon__': op_notify,
    '__operator_GreaterGreater__': op_update,
}




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
#!/usr/bin/env python

import random

from traits.api import HasTraits, Str

from enaml.factory import EnamlFactory
from enaml.color import Color
from enaml.style_sheet import style

colors = Color.color_map.keys()

class Model(HasTraits):
    """ Traits model to interact with the enaml widget

    """

    message = Str('Foo Model Message!')
    window_title = Str('Window Title!')

    def print_msg(self, args):
        """ Print a message to standard output.

        """
        print self.message, args

    def randomize(self, string):
        """ Suffle the letters in the string.

        """
        l = list(string)
        random.shuffle(l)
        return ''.join(l)

    def update_style(self):
        """ Update the style of the widget

        """
        # create a new style name .error_colors
        # with random foreground/background colors
        sty = style('.error_colors',
            color = random.choice(colors),
            background_color=random.choice(colors),
        )
        # update the vew style sheet with the new style
        view.style_sheet.update(sty)


fact = EnamlFactory('./example.enaml')

view = fact(model=Model())

view.show()




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import Property, on_trait_change

from ..core.base_component import BaseComponent


class Inline(BaseComponent):
    """ A BaseComponent subclass which allows children to be statically
    inlined into a parent.

    """
    #: A read-only property which returns the toolkit widget for this
    #: component. This call is proxied to the parent and will return
    #: None if the parent does not have a toolkit widget defined. 
    #: It is necessary to proxy the toolkit_widget so that any child
    #: Include components can access the proper toolkit widget to 
    #: pass down to their dynamic components.
    toolkit_widget = Property

    #--------------------------------------------------------------------------
    # Property Getters
    #--------------------------------------------------------------------------
    def _get_toolkit_widget(self):
        """ The property getter for the 'toolkit_widget' attribute.

        """
        try:
            res = self.parent.toolkit_widget
        except AttributeError:
            res = None
        return res

    #--------------------------------------------------------------------------
    # Overrides
    #--------------------------------------------------------------------------
    def get_actual(self):
        """ A reimplemented parent class method which returns the inlined
        subcomponents for this component.

        """
        return sum([c.get_actual() for c in self._subcomponents], [])

    #--------------------------------------------------------------------------
    # Change Handlers
    #--------------------------------------------------------------------------
    @on_trait_change('_subcomponents:_actual_updated')
    def handle_actual_updated(self):
        """ Handles the '_actual_updated' event being fired on any of the
        subcomponents. It proxies this event up to the parent so that it
        can request the new children and relayout if necessary.

        """
        self._actual_updated()




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import re

from enaml.validation import AbstractValidator


class PhoneNumberValidator(AbstractValidator):
    """ A really dumb phone number validator.

    """
    all_digits = re.compile(r'[0-9]{10}$')

    dashes = re.compile(r'([0-9]{3})\-([0-9]{3})\-([0-9]{4})$')

    proper = re.compile(r'\(([0-9]{3})\)\ ([0-9]{3})\-([0-9]{4})$')

    def _validate(self, text):
        match = self.proper.match(text) or self.dashes.match(text)
        if match:
            area = match.group(1)
            prefix = match.group(2)
            suffix = match.group(3)
            parts = tuple(map(int, (area, prefix, suffix)))
            return (self.ACCEPTABLE, parts)
        match = self.all_digits.match(text)
        if match:
            area = text[:3]
            prefix = text[3:6]
            suffix = text[6:10]
            parts = tuple(map(int, (area, prefix, suffix)))
            return (self.ACCEPTABLE, parts)
        return (self.INTERMEDIATE, None)

    def validate(self, text):
        return self._validate(text)[0]
    
    def convert(self, text):
        return self._validate(text)[1]
    
    def format(self, value):
        area, prefix, suffix = value
        return '(%s) %s-%s' % (area, prefix, suffix)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import numpy as np
from scipy.misc import imread

from skimage.filter import canny
from skimage.transform import probabilistic_hough

from chaco.api import Plot, ArrayPlotData, bone
from enable.api import ColorTrait
from traits.api import (
    HasTraits, File, Array, Property, Instance, Any, Int, List, Tuple, Range,
    DelegatesTo, cached_property, on_trait_change,
)

from segments_overlay import SegmentsOverlay


class ImageProcessor(HasTraits):

    original_image = Array

    canny_sigma = Range(value=1, low=0.0)
    canny_low_threshold = Range(value=0.1, low=0.0)
    canny_high_threshold = Range(value=0.2, low=0.0)
    canny_image = Property(
        Array, depends_on=['original_image', 'canny_sigma', 
                           'canny_low_threshold', 'canny_high_threshold'],
    )

    hough_threshold = Int(10)
    hough_line_length = Int(50)
    hough_line_gap = Int(10)
    hough_segments = Property(
        List, depends_on=['canny_image', 'hough_threshold', 
                          'hough_line_length', 'hough_line_gap'],
    )

    def _original_image_default(self):
        return np.array([[0]])

    @cached_property
    def _get_canny_image(self):
        ci = canny(
            self.original_image, sigma=self.canny_sigma,
            low_threshold=self.canny_low_threshold,
            high_threshold=self.canny_high_threshold,
        )
        return ci

    @cached_property
    def _get_hough_segments(self):
        segs = probabilistic_hough(
            self.canny_image, threshold=self.hough_threshold,
            line_length=self.hough_line_length, line_gap=self.hough_line_gap,
        )
        return segs


class ProcessorModel(HasTraits):

    filename = File

    image = Instance(ImageProcessor, ())
    
    shape = Property(Tuple, depends_on=['image'])
    width = Property(Int, depends_on=['shape'])
    height = Property(Int, depends_on=['shape'])

    original_image = DelegatesTo('image')
    canny_sigma = DelegatesTo('image')
    canny_low_threshold = DelegatesTo('image')
    canny_high_threshold = DelegatesTo('image')
    canny_image = DelegatesTo('image')

    hough_threshold = DelegatesTo('image')
    hough_line_length = DelegatesTo('image')
    hough_line_gap = DelegatesTo('image')
    hough_segments = DelegatesTo('image')

    segments_overlay = Any

    original_alpha = Range(value=1.0, low=0.0, high=1.0) 

    plot_data = Instance(ArrayPlotData)
    main_plot = Instance(Plot)

    canny_color = ColorTrait('red')
    canny_alpha = Range(value=1.0, low=0.0, high=1.0)
    canny_plot_image = Property(
        Array, depends_on=['image', 'canny_image', 'canny_sigma',
                           'canny_low_threshold', 'canny_high_threshold',
                           'canny_color'],
    )

    hough_segments_alpha = Range(value=1.0, low=0.0, high=1.0) 
    hough_segments_color = ColorTrait('green')

    original_plot = Any
    canny_plot = Any

    background_color = ColorTrait('black')

    #--------------------------------------------------------------------------
    # Trait defaults
    #--------------------------------------------------------------------------
    def _plot_data_default(self):
        pd = ArrayPlotData(
            original_image=self.image.original_image,
            canny_plot_image=self.canny_plot_image,
        )
        return pd

    def _main_plot_default(self):
        p = Plot(self.plot_data, default_origin='top left', padding=0)
        self.original_plot = p.img_plot(
            'original_image', colormap=bone, alpha=self.original_alpha,
            bgcolor=self.background_color_,
        )[0]
        self.canny_plot = p.img_plot(
            'canny_plot_image', alpha=self.canny_alpha,
        )[0]
        p.x_axis = None
        p.y_axis = None
        self.segments_overlay = SegmentsOverlay(
            component=self.canny_plot, image_size=self.image.canny_image.shape,
        )
        p.overlays.append(self.segments_overlay)
        return p

    #--------------------------------------------------------------------------
    # Trait property methods
    #--------------------------------------------------------------------------
    @cached_property
    def _get_shape(self):
        shape = self.image.original_image.shape
        return shape

    @cached_property
    def _get_width(self):
        width = self.shape[1]
        return width

    @cached_property
    def _get_height(self):
        height = self.shape[0]
        return height

    @cached_property
    def _get_canny_plot_image(self):
        # self.image.canny_image is an array of bools.
        # self.canny_plot_image is a 3D array (2D image with (r,g,b,a)
        # values).  The color (0,0,0,0) (transparent) is assigned where
        # canny_image is False, and (canny_color_, 255) is assigned where
        # canny_image is True.
        ci = self.image.canny_image
        x = np.zeros(ci.shape + (4,), dtype=np.uint8)
        x[ci, :3] = [255 * c for c in self.canny_color_[:3]]
        x[ci, 3] = 255
        return x

    #--------------------------------------------------------------------------
    # Trait change handlers
    #--------------------------------------------------------------------------
    def _filename_changed(self):
        try:
            image_data = imread(self.filename, flatten=True).astype(np.float32)
            image_data /= image_data.max()
            image = ImageProcessor(original_image=image_data)
        except IOError:
            image = ImageProcessor(original_image=np.array([[0]]))
        self.image = image

        self.plot_data['original_image'] = self.image.original_image
        self.plot_data['canny_plot_image'] = self.canny_plot_image
        self.segments_overlay.image_size = self.image.canny_image.shape
        self.segments_overlay.segments = self.image.hough_segments

    @on_trait_change('canny_plot_image')
    def _canny_parameter_changed(self):
        self.plot_data['canny_plot_image'] = self.canny_plot_image

    def _original_alpha_changed(self):
        self.original_plot.alpha = self.original_alpha
        # Temporary hack...
        self.plot_data['original_image'] = self.image.original_image
        self.plot_data['canny_plot_image'] = self.canny_plot_image

    @on_trait_change('canny_alpha, canny_color')
    def _canny_display_parameters_changed(self):
        self.canny_plot.alpha = self.canny_alpha
        # Temporary hack...
        self.plot_data['original_image'] = self.image.original_image
        self.plot_data['canny_plot_image'] = self.canny_plot_image

    @on_trait_change('hough_segments_alpha, hough_segments_color')
    def _hough_display_parameters_changed(self):
        self.segments_overlay.line_color = self.hough_segments_color
        self.segments_overlay.alpha = self.hough_segments_alpha
        self.segments_overlay.request_redraw()

    def _background_color_changed(self):
        self.original_plot.bgcolor = self.background_color_
        self.original_plot.request_redraw()

    @on_trait_change('hough_segments')
    def _update_hough_segments_plot(self):
        self.segments_overlay.segments = self.image.hough_segments
        self.canny_plot.request_redraw()




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from enaml.backends.qt.qt import QtGui

from .qt_test_assistant import QtTestAssistant

from .. import field


QT_2_ENAML_PASSWORD_MODES = {
    QtGui.QLineEdit.Normal: 'normal',
    QtGui.QLineEdit.Password: 'password',
    QtGui.QLineEdit.NoEcho: 'silent',
}


class TestQtField(QtTestAssistant, field.TestField):
    """ QtField tests. 

    """
    def get_value(self, widget):
        """ Get the visible text of a field.

        """
        return widget.text()

    def edit_text(self, widget, text):
        """ Simulate typing in a field.

        """
        widget.insert(text)
        
    def change_text(self, widget, text):
        """ Change text programmatically, rather than "edit" it.

        """
        widget.setText(text)

    def set_cursor_position(self, widget, index):
        """ Set the cursor at a specific position.

        """
        widget.setCursorPosition(index)

    def get_cursor_position(self, widget):
        """ Get the cursor position.

        """
        return widget.cursorPosition()

    def set_selected_text(self, widget, start, stop):
        """ Select text in a field.

        """
        widget.setSelection(start, stop - start)

    def get_selected_text(self, widget):
        """ Get the currently-selected text from a field.

        """
        return widget.selectedText()

    def get_password_mode(self, widget):
        """ Get the password mode status of the widget

        """
        mode = widget.echoMode()
        return QT_2_ENAML_PASSWORD_MODES[mode]

    def press_return(self, widget):
        """ Simulate a press of the 'Return' key.

        """
        widget.returnPressed.emit()

    def gain_focus_if_needed(self, widget):
        """ Have the widget gain focus if required for the tests.

        """
        # qt does the right thing, so we don't need to gain focus.
        pass




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
""" Driver for the sine plot example.

The window will show a Chaco plot on the left with sin(x) and cos(x) 
functions. The right side of the window will be a Form showing sliders 
that control the phase, amplitude, and angular frequency of the sinusoidal 
functions. The plot will update whenever the slider values change.

This demo requires numpy and Chaco.

"""
import numpy as np

from traits.api import HasTraits, Array, Float, Property

import enaml


class SineModel(HasTraits):
    #: The phase of the sine wave.
    phase = Float()

    #: The amplitude of the sine wave.
    amplitude = Float(1.0)

    #: The angular frequency of the sine wave.
    w = Float(1.0)

    #: The data.
    x = Array()
    def _x_default(self):
        return np.linspace(-6*np.pi, 6*np.pi, 301)

    y_sin = Property(Array, depends_on=['phase', 'amplitude', 'w'])
    def _get_y_sin(self):
        return self.amplitude * np.sin(self.w * self.x + self.phase)

    y_cos = Property(Array, depends_on=['phase', 'amplitude', 'w'])
    def _get_y_cos(self):
        return self.amplitude * np.cos(self.w * self.x + self.phase)


def main():
    from simple_plot import SimplePlot
    SimplePlot.activate()

    with enaml.imports():
        from sine_plot import Main

    model = SineModel()
    window = Main(model=model)
    window.show()


if __name__ == '__main__':
    main()




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
""" A data sampler api for subsampling a data stream.

"""
from abc import ABCMeta, abstractmethod
import numpy as np


class AbstractDataSampler(object):
    """ An abstract base class which defines the api necessary for a
    data sampler.

    """
    __metaclass__ = ABCMeta
    @abstractmethod
    def sample(self, data):
        """ Samples the given data buffer and returns a possibly new
        buffer with subsampled data.

        Parameters
        ----------
        data : np.ndarray
            The buffer of data to be sampled.
        
        Returns
        -------
        result : np.ndarray
            A new buffer of sampled data, or the original buffer if 
            no sampling needs to be performed.

        """
        raise NotImplementedError


class ProbabilitySampler(AbstractDataSampler):
    """ A concrete implementation of AbstractDataSampler which sub 
    samples data based on a uniform probability distribution.

    """
    def __init__(self, probability=1.0):
        """ Initialized a ProbabilitySampler.

        Parameters
        ----------
        probability : float, optional
            A floating point probability in the range(0.0, 1.0)
        
        """
        prob = probability
        assert 0.0 <= prob <= 1.0, "Probability must be 0.0 <= val <= 1.0"
        self.probability = prob 

    def sample(self, data):
        num_entries = data.shape[0]
        coin_flips = np.random.random(num_entries)
        mask = coin_flips < self.probability
        return data[mask]




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import unittest

from enaml.validation import (
    NumberValidator, IntegralNumberValidator, RealNumberValidator,
    ComplexNumberValidator, IntValidator, LongValidator, FloatValidator,
    ComplexValidator, BinValidator, OctValidator, HexValidator,
    LongBinValidator, LongOctValidator, LongHexValidator,
)


@unittest.skip('Skipping Validation Tests')
class TestNumberValidators(unittest.TestCase):

    def test_number_validator(self):
        v = NumberValidator()
        self.assertEqual(v.validate(u'k'), v.INVALID)
        self.assertEqual(v.validate(u'12'), v.ACCEPTABLE)
        self.assertEqual(v.validate(u'-12'), v.ACCEPTABLE)
        self.assertEqual(v.validate(u'1e7'), v.ACCEPTABLE)

    def test_number_out_of_range(self):
        v = NumberValidator(low=10, high=89)
        self.assertEqual(v.validate(u'12'), v.ACCEPTABLE)
        self.assertEqual(v.validate(u'-12'), v.INVALID)
        self.assertEqual(v.validate(u'9'), v.INTERMEDIATE)
        self.assertEqual(v.validate(u'60k'), v.INVALID)
    
    def test_integral_number_validator(self):
        v = IntegralNumberValidator()




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt import QtCore, QtGui
from .qt_control import QtControl

from ...components.push_button import AbstractTkPushButton


class QtPushButton(QtControl, AbstractTkPushButton):
    """ A Qt implementation of PushButton.

    """
    
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying QPushButton control.

        """
        self.widget = QtGui.QPushButton(parent)

    def initialize(self):
        """ Intializes the widget with the attributes of this instance.

        """
        super(QtPushButton, self).initialize()
        self.set_text(self.shell_obj.text)
        self.set_icon(self.shell_obj.icon)
        self.set_icon_size(self.shell_obj.icon_size)

    def bind(self):
        """ Connects the event handlers for the push button.

        """
        super(QtPushButton, self).bind()
        widget = self.widget
        widget.clicked.connect(self.on_clicked)
        widget.pressed.connect(self.on_pressed)
        widget.released.connect(self.on_released)

    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_text_changed(self, text):
        """ The change handler for the 'text' attribute.

        """
        self.set_text(text)
        # If the text of the button changes, the size hint has likely
        # change and the layout system needs to be informed.
        self.shell_obj.size_hint_updated()

    def shell_icon_changed(self, icon):
        """ The change handler for the 'icon' attribute.

        """
        self.set_icon(icon)
        # If the icon of the button changes, the size hint has likely
        # changed and the layout system needs to be informed.
        self.shell_obj.size_hint_updated()

    def shell_icon_size_changed(self, icon_size):
        """ The change handler for the 'icon_size' attribute.

        """
        self.set_icon_size(icon_size)
        # If the icon size of the button changes, the size hint has likely
        # changed and the layout system needs to be informed.
        self.shell_obj.size_hint_updated()

    def on_clicked(self):
        """ The event handler for the button's clicked event.

        """
        shell = self.shell_obj
        shell._down = False
        shell.clicked()

    def on_pressed(self):
        """ The event handlers for the button's pressed event.

        """
        shell = self.shell_obj
        shell._down = True
        shell.pressed()

    def on_released(self):
        """ The event handler for the button's released event.

        """
        shell = self.shell_obj
        if shell._down:
            shell._down = False
            shell.released()

    def set_text(self, text):
        """ Sets the text on the button control.

        """
        self.widget.setText(text)

    def set_icon(self, icon):
        """ Sets the icon on the button control.

        """
        if icon is None:
            qicon = QtGui.QIcon()
        else:
            qicon = icon.as_QIcon()
        self.widget.setIcon(qicon)

    def set_icon_size(self, icon_size):
        """ Sets the icon size on the button control.

        """
        if icon_size is not None and (icon_size.width > 0 
                                      and icon_size.height > 0):
            self.widget.setIconSize(QtCore.QSize(*icon_size))




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from collections import deque

from traits.api import (
    Bool, HasStrictTraits, Instance, List, Property, Str, WeakRef, Dict, 
    Disallow,
)

from .expressions import AbstractExpression
from .toolkit import Toolkit
from .trait_types import (
    EnamlEvent, LazyProperty, UserAttribute, UserEvent, ExpressionTrait,
)


class BaseComponent(HasStrictTraits):
    """ The most base class of the Enaml component hierarchy.

    All declarative Enaml classes should inherit from this class. This 
    class is not meant to be instantiated directly.

    """
    #: A readonly property which returns the current instance of
    #: the component. This allows declarative Enaml components to
    #: access self according to the standard attribute scoping rules.
    self = Property

    #: The parent component of this component. It is stored as a weakref
    #: to mitigate issues with reference cycles. A top-level component's
    #: parent is None.
    parent = WeakRef('BaseComponent')

    #: The list of children for this component. This is a read-only
    #: lazy property that is computed based on the static list of
    #: _subcomponents and the items they return by calling their
    #: 'get_actual' method. This list should not be manipulated by
    #: user code.
    children = LazyProperty(
        List(Instance('BaseComponent')), 
        depends_on='_subcomponents:_actual_updated',
    )

    #: Whether the component has been initialized or not. This will be 
    #: set to True after all of the setup() steps defined here are 
    #: completed. It should not be changed afterwards. This can be used 
    #: to trigger certain actions that need to occur after the component 
    #: has been set up.
    initialized = Bool(False)

    #: An optional name to give to this component to assist in finding
    #: it in the tree. See the 'find_by_name' method.
    name = Str

    #: A reference to the toolkit that was used to create this object.
    toolkit = Instance(Toolkit)

    #: The private dictionary of expression objects that are bound to 
    #: attributes on this component. It should not be manipulated by
    #: user code. Rather, expressions should be bound by calling the 
    #: 'bind_expression' method.
    _expressions = Dict(Str, List(Instance(AbstractExpression)))

    #: The private list of virtual base classes that were used to 
    #: instantiate this component from Enaml source code. The 
    #: EnamlFactory class of the Enaml runtime will directly append
    #: to this list as necessary.
    _bases = List

    #: The private internal list of subcomponents for this component. 
    #: This list should not be manipulated by the user, and should not
    #: be changed after initialization. It can, however, be redefined
    #: by subclasses to limit the type or number of subcomponents.
    _subcomponents = List(Instance('BaseComponent'))

    #: A private event that should be emitted by a component when the 
    #: results of calling get_actual() will result in new values. 
    #: This event is listened to by the parent of subcomponents in order 
    #: to know when to rebuild its list of children. User code will not 
    #: typically interact with this event.
    _actual_updated = EnamlEvent

    #: The HasTraits class defines a class attribute 'set' which is
    #: a deprecated alias for the 'trait_set' method. The problem
    #: is that having that as an attribute interferes with the 
    #: ability of Enaml expressions to resolve the builtin 'set',
    #: since the dynamic attribute scoping takes precedence over
    #: builtins. This resets those ill-effects.
    set = Disallow

    #--------------------------------------------------------------------------
    # Special Methods
    #--------------------------------------------------------------------------
    def __repr__(self):
        """ An overridden repr which returns the repr of the factory 
        from which this component is derived, provided that it is not 
        simply a root constructor. Otherwise, it defaults to the super
        class' repr implementation.

        """
        # If there are any bases, the last one in the list will always 
        # be a constructor. We want to ignore that one and focus on the
        # repr of the virtual base class from which the component was 
        # derived in the Enaml source code.
        bases = self._bases
        if len(bases) >= 2:
            base = bases[0]
            return repr(base)
        return super(BaseComponent, self).__repr__()

    #--------------------------------------------------------------------------
    # Property Getters
    #--------------------------------------------------------------------------
    def _get_self(self):
        """ The property getter for the 'self' attribute.

        """
        return self
        
    def _get_children(self):
        """ The lazy property getter for the 'children' attribute.

        This property getter returns the flattened list of components
        returned by calling 'get_actual()' on each subcomponent.

        """
        return sum([c.get_actual() for c in self._subcomponents], [])
    
    #--------------------------------------------------------------------------
    # Component Manipulation
    #--------------------------------------------------------------------------
    def get_actual(self):
        """ Returns the list of BaseComponent instances which should be
        included as proper children of our parent. By default this 
        simply returns [self]. This method should be reimplemented by 
        subclasses which need to contribute different components to their
        parent's children.

        """
        return [self]
        
    def add_subcomponent(self, component):
        """ Adds the given component as a subcomponent of this object.
        By default, the subcomponent is added to an internal list of 
        subcomponents. This method may be overridden by subclasses to 
        filter or otherwise handle certain subcomponents differently.

        """
        component.parent = self
        self._subcomponents.append(component)
    
    #--------------------------------------------------------------------------
    # Setup Methods 
    #--------------------------------------------------------------------------
    def setup(self, parent=None):
        """ Run the setup process for the ui tree.

        The setup process is fairly complex and involves multiple steps.
        The complexity is required in order to ensure a consistent state
        of the component tree so that default values that are computed 
        from expressions have the necessary information available.

        The setup process is comprised of the following steps:
        
        1)  Abstract objects create their internal toolkit object
        2)  Abstract objects initialize their internal toolkit object
        3)  Bound expression values are explicitly applied
        4)  Abstract objects bind their event handlers
        5)  Abstract objects are added as listeners to the shell object
        6)  Visibility is initialized
        7)  Layout is initialized
        8)  A finalization pass is made
        9)  Nodes are marked as initialized
        
        Many of these setup methods are no-ops, but are defined on this
        BaseComponent for simplicity and continuity. Subclasses that
        need to partake in certain portions of the layout process 
        should re-implement the appropriate setup methods.

        Parameters
        ----------
        parent : native toolkit widget, optional
            If embedding this BaseComponent into a non-Enaml GUI, use 
            this to pass the appropriate toolkit widget that should be 
            the parent toolkit widget for this component.

        """
        self._setup_create_widgets(parent)
        self._setup_init_widgets()
        self._setup_eval_expressions()
        self._setup_bind_widgets()
        self._setup_listeners()
        self._setup_init_visibility()
        self._setup_init_layout()
        self._setup_finalize()
        self._setup_set_initialized()

    def _setup_create_widgets(self, parent):
        """ A setup method that, by default, is a no-op. Subclasses 
        that drive gui toolkit widgets should reimplement this method
        to create the underlying toolkit widget(s).

        """
        for child in self._subcomponents:
            child._setup_create_widgets(parent)

    def _setup_init_widgets(self):
        """ A setup method that, by default, is a no-op. Subclasses 
        that drive gui toolkit widgets should reimplement this method
        to initialize their internal toolkit widget(s).

        """
        for child in self._subcomponents:
            child._setup_init_widgets()

    def _setup_eval_expressions(self):
        """ A setup method that loops over all of bound expressions and
        performs a getattr for those attributes. This ensures that all
        bound attributes are initialized, even if they weren't implicitly
        initialized in any of the previous setup methods.

        """
        for name in self._expressions:
            getattr(self, name)
        for child in self._subcomponents:
            child._setup_eval_expressions()

    def _setup_bind_widgets(self):
        """ A setup method that, by default, is a no-op. Subclasses 
        that drive gui toolkit widgets should reimplement this method
        to bind any event handlers of their internal toolkit widget(s).

        """
        for child in self._subcomponents:
            child._setup_bind_widgets()

    def _setup_listeners(self):
        """ A setup method that, by default, is a no-op. Subclasses 
        that drive gui toolkit widgets should reimplement this method
        to setup an traits listeners necessary to drive their internal
        toolkit widget(s).

        """
        for child in self._subcomponents:
            child._setup_listeners()

    def _setup_init_visibility(self):
        """ A setup method that, by default, is a no-op. Subclasses 
        that drive gui toolkit widgets should reimplement this method
        to initialize the visibility of their widgets.

        """
        for child in self._subcomponents:
            child._setup_init_visibility()

    def _setup_init_layout(self):
        """ A setup method that, by default, is a no-op. Subclasses 
        that manage layout should reimplement this method to initialize
        their underlying layout.

        """
        for child in self._subcomponents:
            child._setup_init_layout()

    def _setup_finalize(self):
        """ A setup method that, by default, is a no-op. Subclasses
        that need to perform process after layout is initialized but
        before a node is marked as fully initialized should reimplement
        this method.

        """
        for child in self._subcomponents:
            child._setup_finalize()

    def _setup_set_initialized(self):
        """ A setup method which updates the initialized attribute of 
        the component to True. This is performed bottom-up.

        """
        for child in self._subcomponents:
            child._setup_set_initialized()
        self.initialized = True

    #--------------------------------------------------------------------------
    # Teardown Methods
    #--------------------------------------------------------------------------
    def destroy(self):
        """ Destroys the component by clearing the list of subcomponents
        and calling 'destroy' on all of the old subcomponents, then gets
        rid of all references to the subcomponents and bound expressions.
        Subclasses that need more control over destruction should 
        reimplement this method.

        """
        for child in self._subcomponents:
            child.destroy()
        del self._subcomponents[:]
        self._expressions.clear()

    #--------------------------------------------------------------------------
    # Layout Stubs
    #--------------------------------------------------------------------------
    def relayout(self):
        """ A method called when the layout of the component's children
        should be refreshed. By default, this method proxies the call up
        the hierarchy until an implementor is found. Any implementors 
        should ensure that the necessary operations take place immediately 
        and are complete before the method returns.

        """
        parent = self.parent
        if parent is not None:
            parent.relayout()

    def request_relayout(self):
        """ A method called when the layout of the component's children
        should be refreshed at some point in the future. By default, this 
        method proxies the call up the hierarchy until an implementor is 
        found. Any implementors should ensure that this method returns 
        immediately, and that relayout occurs at some point in the future.

        """
        parent = self.parent
        if parent is not None:
            parent.request_relayout()

    def refresh(self):
        """ A method called when the positioning of the component's 
        children should be refreshed. By default, this method proxies the 
        call up the hierarchy until an implementor is found. Implementors 
        should ensure that this method takes place immediately, and that
        the refresh is complete before the method returns. 

        Note: This method should perform less work than 'relayout' and 
            should typically only need to be called when the children 
            need to be repositioned, rather than have all of their layout 
            relationships recomputed.

        """
        parent = self.parent
        if parent is not None:
            parent.refresh()
        
    def request_refresh(self):
        """ A method called when the positioning of the component's 
        children should be refreshed at some point in the future. By 
        default, this method proxies the call up the hierarchy until an 
        implementor is found. Implementors should ensure that this method 
        returns immediately, and that the refresh is completed at some 
        time in the future.
        
        Note: This method should perform less work than 'relayout' and 
            should typically only need to be called when the children 
            need to be repositioned, rather than have all of their layout 
            relationships recomputed.

        """
        parent = self.parent
        if parent is not None:
            parent.request_refresh()
        
    def request_relayout_task(self, callback, *args, **kwargs):
        """ Schedule a callback to be executed, followed by a relayout. 
        By default, this method proxies the call up the hierarchy until 
        an implementor is found. Implementors should ensure that the
        callback is executed with given arguments at some point in the
        future and is followed by a relayout. It is suggested that 
        implementors collapse multiple calls to this method which
        results in a single relayout.

        """
        parent = self.parent
        if parent is not None:
            parent.request_relayout_task(callback, *args, **kwargs)
        
    def request_refresh_task(self, callback, *args, **kwargs):
        """ Schedule a callback to be executed, followed by a rerfresh. 
        By default, this method proxies the call up the hierarchy until 
        an implementor is found. Implementors should ensure that the
        callback is executed with given arguments at some point in the
        future and is followed by a relayout. It is suggested that 
        implementors collapse multiple calls to this method which
        results in a single refresh.

        """
        parent = self.parent
        if parent is not None:
            parent.request_refresh_task(callback, *args, **kwargs)

    #--------------------------------------------------------------------------
    # Bound Attribute Handling
    #--------------------------------------------------------------------------
    def add_attribute(self, name, attr_type=object, is_event=False):
        """ Adds an attribute to the base component with the given name
        and ensures that values assigned to this attribute are of a
        given type.

        If the object already has an attribute with the given name,
        an exception will be raised.

        Parameters
        ----------
        name : string
            The name of the attribute to add.
        
        attr_type : type-like object, optional
            An object that behaves like a type for the purposes of a
            call to isinstance. Defaults to object.
        
        is_event : bool, optional
            If True, the added attribute will behave like an event.
            Otherwise, it will behave like a normal attribute. The 
            default is False.

        """
        # Check to see if a trait is already defined. We don't use
        # hasattr here since that might prematurely trigger a trait
        # intialization. We allow overriding traits of type Disallow,
        # UserAttribute, and UserEvent. The first is a consequence of 
        # using HasStrictTraits, where non-existing attributes are 
        # manifested as a Disallow trait. The others allow a custom 
        # derived component to specialize the attribute and event types 
        # of the component from which it is deriving.
        curr = self.trait(name)
        if curr is not None:
            ttype = curr.trait_type
            allowed = (UserAttribute, UserEvent)
            if ttype is not Disallow and not isinstance(ttype, allowed):
                msg = ("Cannot add '%s' attribute. The '%s' attribute on "
                       "the %s object already exists.")
                raise TypeError(msg % (name, name, self))
            
        # At this point we know there are no non-overridable traits 
        # defined for the object, but it is possible that there are 
        # methods or other non-trait attributes using the given name. 
        # We could potentially check for those, but its probably more 
        # useful to allow for overriding such things from Enaml, so we 
        # just go ahead and add the attribute.
        try:
            if is_event:
                self.add_trait(name, UserEvent(attr_type))
            else:
                self.add_trait(name, UserAttribute(attr_type))
        except TypeError:
            msg = ("'%s' is not a valid type for the '%s' attribute "
                   "declaration on %s")
            raise TypeError(msg % (attr_type, name, self))

    def bind_expression(self, name, expression, notify_only=False):
        """ Binds the given expression to the attribute 'name'.
         
        If the attribute does not exist, an exception is raised. A 
        strong reference to the expression object is kept internally.
        If the expression is not notify_only and the object is already
        fully initialized, the value of the expression will be applied
        immediately.

        Parameters
        ----------
        name : string
            The name of the attribute on which to bind the expression.
        
        expression : AbstractExpression
            A concrete implementation of AbstractExpression.
        
        notify_only : bool, optional
            If True, the expression is only a notifier, in which case
            multiple binding is allowed, otherwise the new expression
            overrides any old non-notify expression. Defaults to False.

        """
        curr = self.trait(name)
        if curr is None or curr.trait_type is Disallow:
            msg = "Cannot bind expression. %s object has no attribute '%s'"
            raise AttributeError(msg % (self, name))

        # If this is the first time an expression is being bound to the
        # given attribute, then we hook up a change handler. This ensures
        # that we only get one notification event per bound attribute.
        # We also create the notification entry in the dict, which is 
        # a list with at least one item. The first item will always be
        # the left associative expression (or None) and all following
        # items will be the notify_only expressions.
        expressions = self._expressions
        if name not in expressions:
            self.on_trait_change(self._on_bound_attr_changed, name)
            expressions[name] = [None]

        # There can be multiple notify_only expressions bound to a 
        # single attribute, so they just get appended to the end of
        # the list. Otherwise, the left associative expression gets
        # placed at the zero position of the list, overriding any
        # existing expression.
        if notify_only:
            expressions[name].append(expression)
        else:
            handler = self._on_expression_changed
            old = expressions[name][0]
            if old is not None:
                old.expression_changed.disconnect(handler)
            expression.expression_changed.connect(handler)
            expressions[name][0] = expression
        
            # Hookup support for default value computation.
            if not self.initialized:
                # We only need to add an ExpressionTrait once, since it 
                # will reach back into the _expressions dict as needed
                # and retrieve the bound expression.
                if not isinstance(curr.trait_type, ExpressionTrait):
                    self.add_trait(name, ExpressionTrait(curr))
            else:
                # If the component is already initialized, and the given
                # expression supports evaluation, update the attribute 
                # with the current value.
                val = expression.eval()
                if val is not NotImplemented:
                    setattr(self, name, val)

    def _on_expression_changed(self, expression, name, value):
        """ A private signal callback for the expression_changed signal
        of the bound expressions. It updates the value of the attribute
        with the new value from the expression.

        """
        setattr(self, name, value)
    
    def _on_bound_attr_changed(self, obj, name, old, new):
        """ A private handler which is called when any attribute which
        has a bound signal changes. It calls the notify method on each
        of the expressions bound to that attribute, but only after the
        component has been fully initialized.

        """
        # The check for None is for the case where there are no left 
        # associative expressions bound to the attribute, so the first
        # entry in the list is still None.
        if self.initialized:
            for expr in self._expressions[name]:
                if expr is not None:
                    expr.notify(old, new)

    #--------------------------------------------------------------------------
    # Auxiliary Methods 
    #--------------------------------------------------------------------------
    def when(self, switch):
        """ A method which returns itself or None based on the truth of
        the argument.

        This can be useful to easily turn off the effects of a component
        if various situations such as constraints-based layout.

        Parameters
        ----------
        switch : bool
            A boolean which indicates whether the instance or None 
            should be returned.
        
        Returns
        -------
        result : self or None
            If 'switch' is boolean True, self is returned. Otherwise,
            None is returned.

        """
        if switch:
            return self
    
    def traverse(self, depth_first=False):
        """ Yields all of the nodes in the tree, from this node downward.

        Parameters
        ----------
        depth_first : bool, optional
            If True, yield the nodes in depth first order. If False,
            yield the nodes in breadth first order. Defaults to False.

        """
        if depth_first:
            stack = [self]
            stack_pop = stack.pop
            stack_extend = stack.extend
        else:
            stack = deque([self])
            stack_pop = stack.popleft
            stack_extend = stack.extend

        while stack:
            item = stack_pop()
            yield item
            stack_extend(item.children)
    
    def traverse_ancestors(self, root=None):
        """ Yields all of the nodes in the tree, from the parent of this 
        node updward, stopping at the given root.

        Parameters
        ----------
        root : BaseComponent, optional
            The component at which to stop the traversal. Defaults
            to None

        """
        parent = self.parent
        while parent is not root and parent is not None:
            yield parent
            parent = parent.parent

    def find_by_name(self, name):
        """ Locate and return a named item that exists in the subtree
        which starts at this node.

        This method will traverse the tree of components, breadth first,
        from this point downward, looking for a component with the given
        name. The first one with the given name is returned, or None if
        no component is found.

        Parameters
        ----------
        name : string
            The name of the component for which to search.
        
        Returns
        -------
        result : BaseComponent or None
            The first component found with the given name, or None if 
            no component is found.
        
        """
        for cmpnt in self.traverse():
            if cmpnt.name == name:
                return cmpnt

    def toplevel_component(self):
        """ Walks up the tree of components starting at this node and
        returns the toplevel node, which is the first node encountered
        without a parent.

        """
        cmpnt = self
        while cmpnt is not None:
            res = cmpnt
            cmpnt = cmpnt.parent
        return res




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .wx_widget_component import WXWidgetComponent

from ...components.constraints_widget import AbstractTkConstraintsWidget


class WXConstraintsWidget(WXWidgetComponent, AbstractTkConstraintsWidget):
    """ A Wx implementation of ConstraintsWidget.

    """
    pass

